from typing import Any, Dict

# Functions in here should be ported to monty if practical


def merge_dicts(
    d1: Dict[str, Any],
    d2: Dict[str, Any],
    remove_none: bool = False,
    remove_false: bool = False,
) -> Dict[str, Any]:
    """
    Merges two dictionaries into a single dictionary. If both dictionaries
    have the same key, the value from the second dictionary will be used. This
    is done in a case-insensitive manner.

    Parameters
    ----------
    d1
        First dictionary.
    d2
        Second dictionary, which has priority.
    remove_none
        If True, all keys with a value of None in the merged dictionary will be removed.
    remove_false
        If True, all keys with a value of False in the merged dictionary will be removed.
    """
    d1_clean = {k.lower(): v for k, v in d1.items()}
    d2_clean = {k.lower(): v for k, v in d2.items()}
    d_merged = {**d1_clean, **d2_clean}
    if remove_none:
        d_merged = {k: v for k, v in d_merged.items() if v is not None}
    if remove_false:
        d_merged = {k: v for k, v in d_merged.items() if v is not False}
    return d_merged



import multiprocessing
import os
from pathlib import Path
from shutil import copy

from ase.build import molecule
from jobflow.managers.local import run_locally

from quacc.recipes.orca.core import RelaxMaker, StaticMaker

FILE_DIR = Path(__file__).resolve().parent
ORCA_DIR = os.path.join(FILE_DIR, "orca_run")


def setup_module():
    for f in os.listdir(ORCA_DIR):
        copy(os.path.join(ORCA_DIR, f), os.path.join(os.getcwd(), f))


def teardown_module():
    for f in os.listdir(ORCA_DIR):
        if os.path.exists(os.path.join(os.getcwd(), f)):
            os.remove(os.path.join(os.getcwd(), f))


def test_static_maker():

    atoms = molecule("H2")
    nprocs = multiprocessing.cpu_count()

    job = StaticMaker().make(atoms)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["nsites"] == len(atoms)
    assert output["name"] == "ORCA-Static"
    assert (
        output["parameters"]["orcasimpleinput"]
        == "wb97x-d def2-tzvp sp slowconv normalprint"
    )
    assert output["parameters"]["orcablocks"] == f"%pal nprocs {nprocs} end"
    assert output["parameters"]["charge"] == 0
    assert output["parameters"]["mult"] == 1

    job = StaticMaker(
        input_swaps={"def2-SVP": True, "def2-TZVP": False},
        block_swaps={"%scf maxiter 300 end": True},
    ).make(atoms, charge=-2, mult=3)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["nsites"] == len(atoms)
    assert output["name"] == "ORCA-Static"
    assert output["parameters"]["charge"] == -2
    assert output["parameters"]["mult"] == 3
    assert (
        output["parameters"]["orcasimpleinput"]
        == "wb97x-d sp slowconv normalprint def2-svp"
    )
    assert (
        output["parameters"]["orcablocks"]
        == f"%scf maxiter 300 end %pal nprocs {nprocs} end"
    )


def test_relax_maker():

    atoms = molecule("H2")
    nprocs = multiprocessing.cpu_count()

    job = RelaxMaker().make(atoms)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["nsites"] == len(atoms)
    assert output["parameters"]["charge"] == 0
    assert output["parameters"]["mult"] == 1
    assert output["name"] == "ORCA-Relax"
    assert (
        output["parameters"]["orcasimpleinput"]
        == "wb97x-d def2-tzvp opt slowconv normalprint"
    )
    assert output["parameters"]["orcablocks"] == f"%pal nprocs {nprocs} end"

    job = RelaxMaker(
        input_swaps={
            "HF": True,
            "wb97x-d": False,
            "def2-SVP": True,
            "def2-TZVP": False,
        },
        block_swaps={"%scf maxiter 300 end": True},
    ).make(atoms, charge=-2, mult=3)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["nsites"] == len(atoms)
    assert output["name"] == "ORCA-Relax"
    assert output["parameters"]["charge"] == -2
    assert output["parameters"]["mult"] == 3
    assert (
        output["parameters"]["orcasimpleinput"]
        == "opt slowconv normalprint hf def2-svp"
    )
    assert (
        output["parameters"]["orcablocks"]
        == f"%scf maxiter 300 end %pal nprocs {nprocs} end"
    )



from dataclasses import dataclass
from typing import Any, Dict

from ase.atoms import Atoms
from ase.optimize import FIRE
from ase.optimize.optimize import Optimizer
from jobflow import Maker, job
from monty.dev import requires

try:
    from xtb.ase.calculator import XTB
except ModuleNotFoundError:
    XTB = None
from quacc.schemas.calc import summarize_run


@dataclass
class StaticMaker(Maker):
    """
    Class to carry out a single-point calculation.

    Parameters
    ----------
    name
        Name of the job.
    method
        GFN0-xTB, GFN1-xTB, GFN2-xTB, GFN-FF.
    xtb_kwargs
        Dictionary of custom kwargs for the xTB calculator.
    """

    name: str = "xTB-Static"
    method: str = "GFN2-xTB"
    xtb_kwargs: Dict[str, Any] = None

    @job
    @requires(
        XTB, "xTB-python must be installed. Try conda install -c conda-forge xtb-python"
    )
    def make(self, atoms: Atoms) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object`

        Returns
        -------
        Dict
            Summary of the run.
        """
        xtb_kwargs = self.xtb_kwargs or {}

        atoms.calc = XTB(method=self.method, **xtb_kwargs)
        atoms.get_potential_energy()
        summary = summarize_run(atoms, additional_fields={"name": self.name})

        return summary


@dataclass
class RelaxMaker(Maker):
    """
    Class to relax a structure.

    Parameters
    ----------
    name
        Name of the job.
    method
        GFN0-xTB, GFN1-xTB, GFN2-xTB, GFN-FF.
    xtb_kwargs
        Dictionary of custom kwargs for the xTB calculator.
    optimizer
        .Optimizer class to use for the relaxation.
    fmax
        Tolerance for the force convergence (in eV/A).
    opt_kwargs
        Dictionary of kwargs for the optimizer.
    """

    name: str = "xTB-Relax"
    method: str = "GFN2-xTB"
    xtb_kwargs: Dict[str, Any] = None
    optimizer: Optimizer = FIRE
    fmax: float = 0.03
    opt_kwargs: Dict[str, Any] = None

    @job
    @requires(
        XTB, "xTB-python must be installed. Try conda install -c conda-forge xtb-python"
    )
    def make(self, atoms: Atoms) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object

        Returns
        -------
        Dict
            Summary of the run.
        """
        xtb_kwargs = self.xtb_kwargs or {}
        opt_kwargs = self.opt_kwargs or {}

        # We always want to save the logfile and trajectory, so we will set some default
        # values if not specified by the user (and then remove them from the **opt_kwargs)
        logfile = opt_kwargs.get("logfile", None) or "opt.log"
        trajectory = opt_kwargs.get("trajectory", None) or "opt.traj"
        opt_kwargs.pop("logfile", None)
        opt_kwargs.pop("trajectory", None)

        atoms.calc = XTB(method=self.method, **xtb_kwargs)
        dyn = self.optimizer(
            atoms, logfile=logfile, trajectory=trajectory, **opt_kwargs
        )
        dyn.run(fmax=self.fmax)
        summary = summarize_run(atoms, additional_fields={"name": self.name})

        return summary



import os
import shlex

from custodian import Custodian
from custodian.vasp.handlers import (
    FrozenJobErrorHandler,
    IncorrectSmearingHandler,
    LargeSigmaHandler,
    MeshSymmetryErrorHandler,
    NonConvergingErrorHandler,
    PositiveEnergyErrorHandler,
    PotimErrorHandler,
    ScanMetalHandler,
    StdErrHandler,
    UnconvergedErrorHandler,
    VaspErrorHandler,
    WalltimeHandler,
)
from custodian.vasp.jobs import VaspJob
from custodian.vasp.validators import VaspFilesValidator, VasprunXMLValidator

from quacc.util.yaml import load_yaml_settings


def run_custodian():
    # Adapted from https://github.com/materialsproject/atomate2/blob/main/src/atomate2/vasp/run.py

    # Read in default settings
    if "VASP_CUSTODIAN_SETTINGS" in os.environ:
        settings_path = os.environ["VASP_CUSTODIAN_SETTINGS"]
    else:
        raise OSError("Missing environment variable VASP_CUSTODIAN_SETTINGS.")
    config = load_yaml_settings(settings_path)

    # Handlers for VASP
    handlers = []
    handlers_dict = {
        "VaspErrorHandler": VaspErrorHandler(vtst_fixes=config["vtst_swaps"]),
        "FrozenJobErrorHandler": FrozenJobErrorHandler(),
        "IncorrectSmearingHandler": IncorrectSmearingHandler(),
        "LargeSigmaHandler": LargeSigmaHandler(),
        "MeshSymmetryErrorHandler": MeshSymmetryErrorHandler(),
        "NonConvergingErrorHandler": NonConvergingErrorHandler(),
        "PositiveEnergyErrorHandler": PositiveEnergyErrorHandler(),
        "PotimErrorHandler": PotimErrorHandler(),
        "StdErrHandler": StdErrHandler(),
        "UnconvergedErrorHandler": UnconvergedErrorHandler(),
        "WalltimeHandler": WalltimeHandler(),
        "ScanMetalHandler": ScanMetalHandler(),
    }
    validators_dict = {
        "VaspFilesValidator": VaspFilesValidator(),
        "VasprunXMLValidator": VasprunXMLValidator(),
    }

    handlers = []
    for handler_flag in config["handlers"]:
        if handler_flag not in handlers_dict.keys():
            raise ValueError(f"Unknown VASP error handler: {handler_flag}")
        handlers.append(handlers_dict[handler_flag])

    validators = []
    for validator_flag in config["validators"]:
        if validator_flag not in validators_dict.keys():
            raise ValueError(f"Unknown VASP validator: {validator_flag}")
        validators.append(validators_dict[validator_flag])

    # Populate settings
    parallel_cmd = config.get("vasp_parallel_cmd", "") + " "
    vasp_cmd = parallel_cmd + config.get("vasp_cmd", "vasp_std")
    vasp_gamma_cmd = parallel_cmd + config.get("vasp_gamma_cmd", "vasp_gam")
    max_errors = config.get("max_errors", 5)
    wall_time = config.get("custodian_wall_time", None)
    scratch_dir = config.get("scratch_dir", None)
    vasp_job_kwargs = config.get("vasp_job_kwargs", None)
    custodian_kwargs = config.get("custodian_kwargs", None)

    # Run VASP
    vasp_job_kwargs = {} if vasp_job_kwargs is None else vasp_job_kwargs
    custodian_kwargs = {} if custodian_kwargs is None else custodian_kwargs
    split_vasp_cmd = shlex.split(vasp_cmd)
    split_vasp_gamma_cmd = shlex.split(vasp_gamma_cmd)
    vasp_job_kwargs.update({"gamma_vasp_cmd": split_vasp_gamma_cmd})

    # Run with Custodian
    jobs = [VaspJob(split_vasp_cmd, **vasp_job_kwargs)]

    if wall_time is not None:
        handlers = list(handlers) + [WalltimeHandler(wall_time=wall_time)]

    c = Custodian(
        handlers,
        jobs,
        validators=validators,
        max_errors=max_errors,
        scratch_dir=scratch_dir,
        **custodian_kwargs,
    )

    c.run()


if __name__ == "__main__":
    run_custodian()



from typing import Any, Dict

from ase.atoms import Atoms

from quacc.schemas.atoms import atoms_to_metadata
from quacc.util.atoms import prep_next_run as prep_next_run_


def summarize_run(
    atoms: Atoms, prep_next_run: bool = True, additional_fields: Dict[str, Any] = None
) -> Dict[str, Any]:
    """
    Get tabulated results from an Atoms object and calculator and store them in a database-friendly format.
    This is meant to be compatible with all calculator types.

    Parameters
    ----------
    atoms
        ASE Atoms object following a calculation.
    prep_next_run
        Whether the Atoms object stored in {"atoms": atoms} should be prepared for the next run
        This clears out any attached calculator and moves the final magmoms to the initial magmoms.
    additional_fields
        Additional fields to add to the task document.

    Returns
    -------
    Dict
        Dictionary of tabulated inputs/results

    """
    # Make sure there is a calculator with results
    if not atoms.calc:
        raise ValueError("ASE Atoms object has no attached calculator.")
    if not atoms.calc.results:
        raise ValueError("ASE Atoms object's calculator has no results.")

    if additional_fields is None:
        additional_fields = {}

    # Fetch all tabulated results from the attached calculator
    results = {"results": atoms.calc.results}

    # Get the calculator inputs
    inputs = {"parameters": atoms.calc.parameters}

    # Prepares the Atoms object for the next run by moving the
    # final magmoms to initial, clearing the calculator state,
    # and assigning the resulting Atoms object a unique ID.
    if prep_next_run:
        atoms = prep_next_run_(atoms)

    # Get tabulated properties of the structure itself
    atoms_db = atoms_to_metadata(atoms)

    # Create a dictionary of the inputs/outputs
    task_doc = {**atoms_db, **inputs, **results, **additional_fields}

    return task_doc



import os
from copy import deepcopy
from pathlib import Path

from ase.build import bulk, molecule
from ase.io import read

from quacc.calculators.vasp import SmartVasp
from quacc.util.atoms import (
    check_is_metal,
    get_atoms_id,
    get_highest_block,
    prep_next_run,
)

FILE_DIR = Path(__file__).resolve().parent
ATOMS_MAG = read(os.path.join(FILE_DIR, "..", "calculators", "vasp", "OUTCAR_mag.gz"))
ATOMS_NOMAG = read(
    os.path.join(FILE_DIR, "..", "calculators", "vasp", "OUTCAR_nomag.gz")
)
ATOMS_NOSPIN = read(
    os.path.join(FILE_DIR, "..", "calculators", "vasp", "OUTCAR_nospin.gz")
)


def test_get_atoms_id():
    atoms = bulk("Cu")
    md5hash = "d4859270a1a67083343bec0ab783f774"
    assert get_atoms_id(atoms) == md5hash

    atoms.info["test"] = "hi"
    assert get_atoms_id(atoms) == md5hash

    atoms.set_initial_magnetic_moments([1.0])
    md5maghash = "7d456a48c235e05cf17da4abcc433a4f"
    assert get_atoms_id(atoms) == md5maghash


def test_prep_next_run():
    atoms = bulk("Cu")
    md5hash = "d4859270a1a67083343bec0ab783f774"
    atoms = prep_next_run(atoms)
    assert atoms.info.get("_id", None) == md5hash
    assert atoms.info.get("_old_ids", None) is None
    atoms = prep_next_run(atoms)
    assert atoms.info.get("_id", None) == md5hash
    assert atoms.info.get("_old_ids", None) == [md5hash]
    atoms[0].symbol = "Pt"
    new_md5hash = "52087d50a909572d58e01cfb49d4911b"
    atoms = prep_next_run(atoms)
    assert atoms.info.get("_old_ids", None) == [
        md5hash,
        md5hash,
    ]
    assert atoms.info.get("_id", None) == new_md5hash

    atoms = deepcopy(ATOMS_MAG)
    atoms.info["test"] = "hi"
    mag = atoms.get_magnetic_moment()
    init_mags = atoms.get_initial_magnetic_moments()
    mags = atoms.get_magnetic_moments()
    atoms = prep_next_run(atoms)
    assert atoms.info["test"] == "hi"
    assert atoms.calc == None
    assert atoms.get_initial_magnetic_moments().tolist() == mags.tolist()

    atoms = deepcopy(ATOMS_MAG)
    atoms.info["test"] = "hi"
    atoms = prep_next_run(atoms, store_results=True)
    assert atoms.info.get("test", None) == "hi"
    assert atoms.info.get("results", None) is not None
    assert atoms.info["results"].get("calc0", None) is not None
    assert atoms.info["results"]["calc0"]["magmom"] == mag
    atoms = SmartVasp(atoms)
    atoms.calc.results = {"magmom": mag - 2}
    atoms = prep_next_run(atoms, store_results=True)
    assert atoms.info.get("results", None) is not None
    assert atoms.info["results"].get("calc1", None) is not None
    assert atoms.info["results"]["calc0"]["magmom"] == mag
    assert atoms.info["results"]["calc1"]["magmom"] == mag - 2

    atoms = deepcopy(ATOMS_MAG)
    atoms = prep_next_run(atoms, move_magmoms=False)
    assert atoms.get_initial_magnetic_moments().tolist() == init_mags.tolist()

    atoms = deepcopy(ATOMS_NOMAG)
    mag = atoms.get_magnetic_moment()
    atoms = prep_next_run(atoms, store_results=True)
    assert atoms.info.get("results", None) is not None
    assert atoms.info["results"].get("calc0", None) is not None
    assert atoms.info["results"]["calc0"]["magmom"] == mag
    atoms = SmartVasp(atoms)
    atoms.calc.results = {"magmom": mag - 2}
    atoms = prep_next_run(atoms, store_results=True)
    assert atoms.info.get("results", None) is not None
    assert atoms.info["results"].get("calc1", None) is not None
    assert atoms.info["results"]["calc0"]["magmom"] == mag
    assert atoms.info["results"]["calc1"]["magmom"] == mag - 2

    atoms = deepcopy(ATOMS_NOSPIN)
    atoms = prep_next_run(atoms, store_results=True)
    assert atoms.info.get("results", None) is not None
    assert atoms.info["results"].get("calc0", None) is not None
    assert atoms.info["results"]["calc0"].get("magmom", None) is None
    atoms = SmartVasp(atoms)
    atoms.calc.results = {"magmom": mag - 2}
    atoms = prep_next_run(atoms, store_results=True)
    assert atoms.info.get("results", None) is not None
    assert atoms.info["results"].get("calc1", None) is not None
    assert atoms.info["results"]["calc0"].get("magmom", None) is None
    assert atoms.info["results"]["calc1"]["magmom"] == mag - 2


def test_check_is_metal():
    atoms = bulk("Cu")
    assert check_is_metal(atoms) == True
    atoms = bulk("Cu") * (2, 2, 2)
    atoms[-1].symbol = "O"
    assert check_is_metal(atoms) == False
    atoms = molecule("H2O")
    assert check_is_metal(atoms) == False


def test_get_highest_block():
    atoms = bulk("Cu")
    assert get_highest_block(atoms) == "d"
    atoms = bulk("Cu") * (2, 2, 2)
    atoms[-1].symbol = "U"
    assert get_highest_block(atoms) == "f"
    atoms = molecule("H2O")
    assert get_highest_block(atoms) == "p"



import os
from copy import deepcopy
from shutil import copy, rmtree
from tempfile import mkdtemp

from ase.atoms import Atoms
from monty.shutil import copy_r, gzip_dir


def run_calc(
    atoms: Atoms, store_dir: str = None, scratch_dir: str = None, gzip: bool = False
) -> float:
    """
    Run a calculation in a scratch directory and copy the results back to the
    original directory. This can be useful if file I/O is slow in the working
    directory, so long as file transfer speeds are reasonable.

    This is a wrapper around atoms.get_potential_energy().

    Parameters
    ----------
    atoms : .Atoms
        The Atoms object to run the calculation on.
    store_dir : str
        Path where calculation results should be stored. Also will copy all files
        from this directory at runtime. If None, the current working directory will be used.
    scratch_dir : str
        Path to the base directory where a tmp directory will be made for
        scratch files. If None, a temporary directory in $SCRATCH will be used.
        If $SCRATCH is not present, a tmp directory will be made in store_dir.
    gzip : bool
        Whether to gzip the output files.

    Returns
    -------
    .Atoms
        The updated .Atoms object,
    """

    atoms = deepcopy(atoms)
    if atoms.calc is None:
        raise ValueError("Atoms object must have attached calculator.")

    # Find the relevant paths
    if not store_dir:
        store_dir = os.getcwd()
    if not scratch_dir:
        if "SCRATCH" in os.environ:
            scratch_dir = os.path.expandvars("$SCRATCH")
        else:
            scratch_dir = store_dir
    if not os.path.exists(scratch_dir):
        raise OSError(f"Cannot find {scratch_dir}")

    tmp_path = mkdtemp(dir=scratch_dir, prefix="quacc-")

    # Copy files to scratch
    for f in os.listdir(store_dir):
        if os.path.isfile(os.path.join(store_dir, f)):
            copy(os.path.join(store_dir, f), os.path.join(tmp_path, f))

    # Leave a note in the run directory for where the tmp is located in case
    # the job dies partway through
    tmp_path_note = os.path.join(store_dir, "tmp_path.txt")
    with open(tmp_path_note, "w") as f:
        f.write(tmp_path)

    # Run calculation via get_potential_energy()
    atoms.calc.directory = tmp_path
    atoms.get_potential_energy()

    # gzip files and recursively copy files back to store_dir
    if gzip:
        gzip_dir(tmp_path)
    copy_r(tmp_path, store_dir)

    # Remove the scratch note and tmp dir
    if os.path.exists(tmp_path_note):
        os.remove(tmp_path_note)
    if os.path.exists(tmp_path):
        rmtree(tmp_path)

    return atoms



import os
from pathlib import Path

import pytest
from ase.build import bulk
from ase.io import read
from monty.json import MontyDecoder, jsanitize

from quacc.calculators.vasp import SmartVasp
from quacc.schemas.cclib import summarize_run

FILE_DIR = Path(__file__).resolve().parent

run1 = os.path.join(FILE_DIR, "gaussian_run1")
log1 = os.path.join(run1, "gautest.log")


def test_summarize_run():

    # Make sure metadata is made
    atoms = read(log1)
    results = summarize_run(atoms, ".log", dir_path=run1)
    assert results["nsites"] == len(atoms)
    assert results["atoms"] == atoms
    assert results["spin_multiplicity"] == 1
    assert results["nsites"] == 6
    assert results["metadata"].get("success", None) == True

    # Make sure default dir works
    cwd = os.getcwd()
    os.chdir(run1)
    summarize_run(atoms, ".log")
    os.chdir(cwd)

    # Make sure info tags are handled appropriately
    atoms = read(log1)
    atoms.info["test_dict"] = {"hi": "there", "foo": "bar"}
    results = summarize_run(atoms, ".log", dir_path=run1)
    assert atoms.info.get("test_dict", None) == {"hi": "there", "foo": "bar"}
    assert results.get("atoms_info", {}) != {}
    assert results["atoms_info"].get("test_dict", None) == {"hi": "there", "foo": "bar"}
    assert results["atoms"].info.get("test_dict", None) == {"hi": "there", "foo": "bar"}

    # Make sure magnetic moments are handled appropriately
    atoms = read(os.path.join(run1, log1))
    atoms.set_initial_magnetic_moments([3.14] * len(atoms))
    atoms.calc.results["magmoms"] = [2.0] * len(atoms)
    results = summarize_run(atoms, ".log", dir_path=run1)

    assert atoms.calc is not None
    assert atoms.get_initial_magnetic_moments().tolist() == [3.14] * len(atoms)

    assert results["atoms"].get_initial_magnetic_moments().tolist() == [2.0] * len(
        atoms
    )
    assert results["atoms"].calc is None

    # Make sure Atoms magmoms were not moved if specified
    atoms = read(os.path.join(run1, log1))
    atoms.set_initial_magnetic_moments([3.14] * len(atoms))
    results = summarize_run(atoms, ".log", dir_path=run1, prep_next_run=False)
    assert atoms.get_initial_magnetic_moments().tolist() == [3.14] * len(atoms)
    assert results["atoms"].get_initial_magnetic_moments().tolist() == [3.14] * len(
        atoms
    )

    # test document can be jsanitized and decoded
    d = jsanitize(results, strict=True, enum_values=True)
    MontyDecoder().process_decoded(d)


def test_errors():
    atoms = bulk("Cu")
    with pytest.raises(ValueError):
        summarize_run(atoms, ".log", dir_path=run1)

    atoms = SmartVasp(atoms)
    with pytest.raises(ValueError):
        summarize_run(atoms, ".log", dir_path=run1)



import os

import pytest
from ase.build import bulk, molecule
from jobflow.managers.local import run_locally

try:
    from xtb.ase.calculator import XTB
except ModuleNotFoundError:
    XTB = None
from quacc.recipes.xtb.core import RelaxMaker, StaticMaker


def teardown_module():
    for f in os.listdir("."):
        if (
            f.endswith(".log")
            or f.endswith(".pckl")
            or f.endswith(".traj")
            or f == "gfnff_topo"
        ):
            os.remove(f)


@pytest.mark.skipif(
    XTB is None,
    reason="xTB-python must be installed. Try conda install -c conda-forge xtb-python",
)
def test_static_maker():

    atoms = molecule("H2O")
    job = StaticMaker().make(atoms)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["spin_multiplicity"] == 1
    assert output["nsites"] == len(atoms)
    assert output["parameters"]["method"] == "GFN2-xTB"
    assert output["name"] == "xTB-Static"
    assert output["results"]["energy"] == pytest.approx(-137.96777587302995)

    job = StaticMaker(method="GFN1-xTB").make(atoms)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["parameters"]["method"] == "GFN1-xTB"
    assert output["results"]["energy"] == pytest.approx(-156.96750578831137)

    job = StaticMaker(method="GFN-FF").make(atoms)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["parameters"]["method"] == "GFN-FF"
    assert output["results"]["energy"] == pytest.approx(-8.912667188932252)

    atoms = bulk("Cu")
    job = StaticMaker(method="GFN1-xTB").make(atoms)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["results"]["energy"] == pytest.approx(-119.77643232313169)


@pytest.mark.skipif(
    XTB is None,
    reason="xTB-python must be installed. Try conda install -c conda-forge xtb-python",
)
def test_relax_maker():
    atoms = molecule("H2O")
    job = RelaxMaker().make(atoms)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["spin_multiplicity"] == 1
    assert output["nsites"] == len(atoms)
    assert output["parameters"]["method"] == "GFN2-xTB"
    assert output["name"] == "xTB-Relax"
    assert output["results"]["energy"] == pytest.approx(-137.9764670127011)

    job = RelaxMaker(method="GFN1-xTB").make(atoms)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["parameters"]["method"] == "GFN1-xTB"
    assert output["results"]["energy"] == pytest.approx(-156.9763496338962)

    job = RelaxMaker(method="GFN-FF").make(atoms)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["parameters"]["method"] == "GFN-FF"
    assert output["results"]["energy"] == pytest.approx(-8.915963671549369)

    atoms = bulk("Cu")
    job = RelaxMaker(method="GFN1-xTB").make(atoms)
    responses = run_locally(job, ensure_success=True)
    output = responses[job.uuid][1].output
    assert output["results"]["energy"] == pytest.approx(-119.77643232313169)



import numpy as np
import pytest
from ase.atoms import Atoms

from quacc.schemas.calc import summarize_run as calc_summarize_run


def mock_get_potential_energy(self, **kwargs):
    # Instead of running .get_potential_energy(), we mock it by attaching
    # dummy results to the atoms object and returning a fake energy. This
    # works because the real atoms.get_potential_energy() takes one argument
    # (i.e. self) and that self object is the atoms object.
    e = -1.0
    self.calc.results = {
        "energy": e,
        "magmom": 0.0,
        "magmoms": [0.0] * len(self),
        "free_energy": -1.0,
        "forces": np.array([[0.0, 0.0, 0.0]] * len(self)),
    }
    return e


@pytest.fixture(autouse=True)
def patch_get_potential_energy(monkeypatch):
    # Monkeypatch the .get_potential_energy() method of the .Atoms object so
    # we aren't running the actual calculation during testing.
    monkeypatch.setattr(Atoms, "get_potential_energy", mock_get_potential_energy)


def mock_summarize_run(atoms, **kwargs):
    # Instead of running the VASP-specific summarize_run(), we mock it with the
    # general calculator schema which does not require VASP files to be
    # in the working directory and will work with pytest.

    prep_next_run = kwargs.get("prep_next_run", True)
    additional_fields = kwargs.get("additional_fields", None)
    output = calc_summarize_run(
        atoms, prep_next_run=prep_next_run, additional_fields=additional_fields
    )
    return output


@pytest.fixture(autouse=True)
def patch_summarize_run(monkeypatch):
    # Monkeypatch the summarize_run() function so that we aren't relying on real
    # VASP files to be in the working directory during the test. Note that even though
    # summarize_run() is a function in the quacc.schemas.vasp module, we modify it
    # only in quacc.recipes.vasp.core/.slabs because otherwise it will not work properly.
    monkeypatch.setattr("quacc.recipes.vasp.core.summarize_run", mock_summarize_run)
    monkeypatch.setattr("quacc.recipes.vasp.slabs.summarize_run", mock_summarize_run)



import os

import pytest

from quacc.util.pop_analysis import run_bader, run_chargemol


def mock_bader_analysis(*args, **kwargs):
    bader_stats = {
        "min_dist": [1.0],
        "charge": [1.0],
        "atomic_volume": [1.0],
        "vacuum_charge": 1.0,
        "vacuum_volume": 1.0,
        "bader_version": 1.0,
        "reference_used": [0.0],
        "charge_transfer": [-1.0],
        "magmom": [0.0],
    }
    return bader_stats


def mock_chargemol_analysis(*args, **kwargs):
    chargemol_stats = {
        "ddec": {
            "partial_charges": [1.0],
            "spin_moments": [0.0],
            "dipoles": [1.0],
            "rsquared_moments": [1.0],
            "rcubed_moments": [1.0],
            "rfourth_moments": [1.0],
        },
        "cm5": {
            "partial_charges": [1.0],
        },
    }
    return chargemol_stats


@pytest.fixture(autouse=True)
def patch_pop_analyses(monkeypatch):
    # Monkeypatch the Bader and Chargemol analyses so they don't run via pytest
    monkeypatch.setattr(
        "quacc.util.pop_analysis.bader_analysis_from_path",
        mock_bader_analysis,
    )
    monkeypatch.setattr(
        "quacc.util.pop_analysis.ChargemolAnalysis",
        mock_chargemol_analysis,
    )


def setup_module():
    for f in ["CHGCAR", "AECCAR0.gz", "AECCAR2.gz", "POTCAR"]:
        with open(f, "w") as w:
            w.write("test")


def teardown_module():
    for f in ["CHGCAR", "AECCAR0.gz", "AECCAR2.gz", "POTCAR"]:
        if os.path.isfile(f):
            os.remove(f)


def test_run_bader():
    bader_stats = run_bader()
    assert bader_stats["min_dist"] == [1.0]
    assert bader_stats["partial_charges"] == [1.0]
    assert bader_stats["spin_moments"] == [0.0]
    assert bader_stats.get("charge") is None
    assert bader_stats.get("charge_transfer") is None
    assert bader_stats.get("reference_used") is None
    assert bader_stats.get("magmom") is None


def test_bader_erorr():
    os.remove("CHGCAR")
    with pytest.raises(FileNotFoundError):
        run_bader()
    with open("CHGCAR", "w") as w:
        w.write("test")


def test_run_chargemol():
    chargemol_stats = run_chargemol(path=".", atomic_densities_path=".")
    assert chargemol_stats["ddec"]["partial_charges"] == [1.0]
    assert chargemol_stats["ddec"]["spin_moments"] == [0.0]


def test_chargemol_erorr():
    with pytest.raises(OSError):
        run_chargemol()

    os.remove("CHGCAR")
    with pytest.raises(FileNotFoundError):
        run_chargemol(atomic_densities_path=".")
    with open("CHGCAR", "w") as w:
        w.write("test")



import hashlib
import os
from copy import deepcopy
from typing import Dict, Optional

import numpy as np
from ase.atoms import Atoms
from ase.io.jsonio import encode
from pymatgen.io.ase import AseAtomsAdaptor

# NOTES:
# - Anytime an Atoms object is converted to a pmg structure, make sure
# to reattach any .info flags to the Atoms object, e.g. via `new_atoms.info = atoms.info.copy()``.
# Note that atoms.info is mutable, so copy it!
# - All major functions should take in Atoms by default and reutrn Atoms
# by default. Pymatgen structures can be returned with an optional kwarg.
# - If you modify the properties of an input Atoms object in any way, make sure to do so
# on a deepcopy because Atoms objects are mutable.


def prep_next_run(
    atoms: Atoms,
    assign_id: bool = True,
    move_magmoms: bool = True,
    store_results: bool = False,
) -> Atoms:
    """
    Prepares the Atoms object for a new run.

    Depending on the arguments, this function will:
        - Move the converged magnetic moments to the initial magnetic moments.
        - Assign a unique ID to the Atoms object in atoms.info["_id"]. Any existing IDs will
        be moved to atoms.info["_old_ids"].
        - Store the calculator results in atoms.info["results"] for later retrieval.
        This makes it so the calculator results are not lost between
        serialize/deserialize cycles, if desired. Each one will be stored in
        atoms.info["results"] = {"calc0": {}, "calc1": {}, ...} with higher numbers
        being the most recent.

    In all cases, the calculator will be reset so new jobs can be run.

    Parameters
    ----------
    atoms
        Atoms object
    assign_id
        Whether to assign a unique ID to the Atoms object in atoms.info["_id"].
        Any existing IDs will be moved to atoms.info["_old_ids"].
    move_magmoms
        If True, move atoms.get_magnetic_moments() to atoms.get_initial_magnetic_moments()
    store_results
        If True, store calculator results in atoms.info["results"]. This makes it so the calculator results
        are not lost between serialize/deserialize cycles, if desired. Each one will be stored in
        atoms.info["results"] = {"calc0": {}, "calc1": {}, ...} with higher numbers being the most recent.

    Returns
    -------
    .Atoms
        Atoms object with calculator results attached in atoms.info["results"]
    """
    try:
        atoms = deepcopy(atoms)
    except TypeError:
        # Workaround for xTB
        atoms = atoms.copy()

    if hasattr(atoms, "calc") and getattr(atoms.calc, "results", None) is not None:

        if store_results:
            # Dump calculator results into the .info tag
            atoms.calc.results["rundir"] = os.getcwd()
            if atoms.info.get("results", None) is None:
                prior_calcs = 0
                atoms.info["results"] = {}
            else:
                prior_calcs = len(atoms.info["results"])

            atoms.info["results"][f"calc{prior_calcs}"] = atoms.calc.results

        # Move converged magmoms to initial magmoms
        # If none were present, then initial magmoms should be set to 0's
        # because a spin-unpolarized calculation was carried out
        if move_magmoms:
            atoms.set_initial_magnetic_moments(
                atoms.calc.results.get("magmoms", [0.0] * len(atoms))
            )

    # Clear off the calculator so we can run a new job. If we don't do this,
    # then something like atoms *= (2,2,2) stil has a calculator attached, which
    # is a bit confusing.
    atoms.calc = None

    # Give the Atoms object a unique ID. This will be helpful for querying later.
    # Also store any old IDs somewhere else for future reference.
    # Note: Kep this at the end of the function so that the ID is assigned based
    # on the returned Atoms object.
    if assign_id:
        if atoms.info.get("_id", None) is not None:
            if atoms.info.get("_old_ids") is None:
                atoms.info["_old_ids"] = []
            atoms.info["_old_ids"].append(atoms.info["_id"])
        atoms.info["_id"] = get_atoms_id(atoms)

    return atoms


def set_magmoms(
    atoms: Atoms,
    elemental_mags_dict: Optional[Dict] = None,
    copy_magmoms: bool = True,
    mag_default: Optional[float] = 1.0,
    mag_cutoff: float = 0.05,
) -> Atoms:
    """
    Sets the initial magnetic moments in the Atoms object.

    This function deserves particular attention. The following logic is applied:
    - If there is a converged set of magnetic moments, those are moved to the
    initial magmoms if copy_magmoms is True.
    - If there is no converged set of magnetic moments but the user has set initial magmoms,
    those are simply used as is.
    - If there are no converged magnetic moments or initial magnetic moments, then
    the default magnetic moments from the preset (if specified) are set as the
    initial magnetic moments.
    - For any of the above scenarios, if mag_cutoff is not None, the newly set
    initial magnetic moments are checked. If all have a magnitude below mag_cutoff,
    then they are all set to 0 (no spin polarization).

    Parameters
    ----------
    atoms
        .Atoms object
    elemental_mags_dict
        Dictionary of elemental symbols and their corresponding magnetic moments to set.
        If None, no default values will be used.
    copy_magmoms
        Whether to copy the magnetic moments from the converged set of magnetic moments to the initial magnetic moments.
    mag_default
        Default magnetic moment to use if no magnetic moments are specified in the preset.
    mag_cutoff
        Magnitude below which the magnetic moments are considered to be zero. If None, no cutoff will be applied

    Returns
    -------
    .Atoms
        Atoms object
    """
    atoms = deepcopy(atoms)

    # Handle the magnetic moments
    # Check if a prior job was run and pull the prior magmoms
    if hasattr(atoms, "calc") and getattr(atoms.calc, "results", None) is not None:
        mags = atoms.calc.results.get("magmoms", [0.0] * len(atoms))
        # Note: It is important that we set mags to 0.0 here rather than None if the
        # calculator has no magmoms because: 1) ispin=1 might be set, and 2) we do
        # not want the preset magmoms to be used.
    else:
        mags = None

    # Check if the user has set any initial magmoms
    has_initial_mags = atoms.has("initial_magmoms")

    # If there are no initial magmoms set and this is not a follow-up job,
    # we may need to add some from the preset yaml.
    if mags is None:
        if not has_initial_mags:

            # If the preset dictionary has default magmoms, set
            # those by element. If the element isn't in the magmoms dict
            # then set it to mag_default.
            if elemental_mags_dict:
                initial_mags = np.array(
                    [
                        elemental_mags_dict.get(atom.symbol, mag_default)
                        for atom in atoms
                    ]
                )
                atoms.set_initial_magnetic_moments(initial_mags)
    # Copy converged magmoms to input magmoms, if copy_magmoms is True
    else:
        if copy_magmoms:
            atoms.set_initial_magnetic_moments(mags)

    # If all the set mags are below mag_cutoff, set them to 0
    if mag_cutoff:
        has_new_initial_mags = atoms.has("initial_magmoms")
        new_initial_mags = atoms.get_initial_magnetic_moments()
        if has_new_initial_mags and np.all(np.abs(new_initial_mags) < mag_cutoff):
            atoms.set_initial_magnetic_moments([0.0] * len(atoms))

    return atoms


def get_atoms_id(atoms: Atoms) -> str:
    """
    Returns a unique ID for the Atoms object. Note: The .info dict
    and calculator is excluded from the hash generation.

    Parameters
    ----------
    atoms
        .Atoms object

    Returns
    -------
    md5hash
        MD5 hash of the .Atoms object
    """

    atoms = deepcopy(atoms)
    atoms.info = {}
    atoms.calc = None
    encoded_atoms = encode(atoms)
    # This is a hack to avoid int32/int64 and float32/float64 differences
    # between machines.
    encoded_atoms = (
        encoded_atoms.replace("int64", "int")
        .replace("int32", "int")
        .replace("float64", "float")
        .replace("float32", "float")
    )
    md5hash = hashlib.md5(encoded_atoms.encode("utf-8")).hexdigest()
    return md5hash


def check_is_metal(atoms: Atoms) -> bool:
    """
    Checks if a structure is a likely metal.

    Parameters
    ----------
    atoms
        .Atoms object

    Returns
    -------
    bool
        True if the structure is likely a metal; False otherwise
    """
    if np.all(atoms.pbc) == False:
        struct = AseAtomsAdaptor.get_molecule(atoms)
    else:
        struct = AseAtomsAdaptor.get_structure(atoms)
    is_metal = all(k.is_metal for k in struct.composition.keys())
    return is_metal


def get_highest_block(atoms: Atoms) -> str:
    """
    Get the highest block (e.g. p-block, d-block f-block) of a structure

    Parameters
    ----------
    atoms
        .Atoms object

    Returns
    -------
    str
        highest block of the structure
    """
    if np.all(atoms.pbc) == False:
        struct = AseAtomsAdaptor.get_molecule(atoms)
    else:
        struct = AseAtomsAdaptor.get_structure(atoms)
    blocks = [site.specie.block for site in struct]
    if "f" in blocks:
        max_block = "f"
    elif "d" in blocks:
        max_block = "d"
    elif "p" in blocks:
        max_block = "p"
    else:
        max_block = "s"

    return max_block



import os
from typing import Any, Dict, List

from ase.atoms import Atoms
from atomate2.common.schemas.cclib import TaskDocument

from quacc.schemas.atoms import atoms_to_metadata
from quacc.util.atoms import prep_next_run as prep_next_run_


def summarize_run(
    atoms: Atoms,
    logfile_extensions: str | List[str],
    dir_path: str = None,
    prep_next_run: bool = True,
    additional_fields: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """
    Get tabulated results from a molecular DFT run and store them in a database-friendly format.
    This is meant to be a general parser built on top of cclib.

    Parameters
    ----------
    atoms
        ASE Atoms object following a calculation.
    logfile_extensions
        Possible extensions of the log file (e.g. ".log", ".out", ".txt", ".chk"). Note that
        only a partial match is needed. For instance, `.log` will match `.log.gz` and `.log.1.gz`.
        If multiple files with this extension are found, the one with the most recent change time
        will be used. For an exact match only, put in the full file name.
    dir_name
        The path to the folder containing the calculation outputs. A value of None specifies the
        current working directory.
    prep_next_run
        Whether the Atoms object storeed in {"atoms": atoms} should be prepared for the next run.
        This clears out any attached calculator and moves the final magmoms to the initial magmoms.
    additional_fields
        Additional fields to add to the task document.

    Returns
    -------
    Dict
        Dictionary of tabulated inputs/results
    """
    # Make sure there is a calculator with results
    if not atoms.calc:
        raise ValueError("ASE Atoms object has no attached calculator.")
    if not atoms.calc.results:
        raise ValueError("ASE Atoms object's calculator has no results.")

    if additional_fields is None:
        additional_fields = {}
    if dir_path is None:
        dir_path = os.getcwd()

    # Fortunately, there is alreayd a cclib parser in Atomate2
    results = TaskDocument.from_logfile(dir_path, logfile_extensions).dict()

    # Remove some key/vals we don't actually ever use
    unused_props = (
        "task_label",
        "tags",
        "molecule",
    )
    for unused_prop in unused_props:
        results.pop(unused_prop, None)

    # Get the calculator inputs
    inputs = {"parameters": atoms.calc.parameters}

    # Prepares the Atoms object for the next run by moving the
    # final magmoms to initial, clearing the calculator state,
    # and assigning the resulting Atoms object a unique ID.
    if prep_next_run:
        atoms = prep_next_run_(atoms)

    # Get tabulated properties of the structure itself
    atoms_db = atoms_to_metadata(atoms)

    # Create a dictionary of the inputs/outputs
    task_doc = {**atoms_db, **inputs, **results, **additional_fields}

    return task_doc



from dataclasses import dataclass
from typing import Any, Dict

from ase.atoms import Atoms
from jobflow import Maker, job

from quacc.calculators.vasp import SmartVasp
from quacc.schemas.vasp import summarize_run
from quacc.util.basics import merge_dicts
from quacc.util.calc import run_calc


@dataclass
class StaticMaker(Maker):
    """
    Class to carry out a single-point calculation.

    Parameters
    ----------
    name
        Name of the job.
    preset
        Preset to use.
    swaps
        Dictionary of custom kwargs for the calculator.
    """

    name: str = "VASP-Static"
    preset: str = None
    swaps: Dict[str, Any] = None

    @job
    def make(self, atoms: Atoms) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object

        Returns
        -------
        Dict
            Summary of the run.
        """
        swaps = self.swaps or {}
        defaults = {
            "ediff": 1e-6,
            "ismear": -5,
            "isym": 2,
            "laechg": True,
            "lcharg": True,
            "lwave": True,
            "nedos": 5001,
            "nsw": 0,
            "sigma": 0.05,
        }
        flags = merge_dicts(defaults, swaps, remove_none=True)

        atoms = SmartVasp(atoms, preset=self.preset, **flags)
        atoms = run_calc(atoms)
        summary = summarize_run(atoms, additional_fields={"name": self.name})

        return summary


@dataclass
class RelaxMaker(Maker):
    """
    Class to relax a structure.

    Parameters
    ----------
    name
        Name of the job.
    preset
        Preset to use.
    volume_relax
        True if a volume relaxation (ISIF = 3) should be performed.
        False if only the positions (ISIF = 2) should be updated.
    swaps
        Dictionary of custom kwargs for the calculator.
    """

    name: str = "VASP-Relax"
    preset: str = None
    volume_relax: bool = True
    swaps: Dict[str, Any] = None

    @job
    def make(self, atoms: Atoms) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object

        Returns
        -------
        Dict
            Summary of the run.
        """
        swaps = self.swaps or {}
        defaults = {
            "ediff": 1e-5,
            "ediffg": -0.02,
            "isif": 3 if self.volume_relax else 2,
            "ibrion": 2,
            "ismear": 0,
            "isym": 0,
            "lcharg": False,
            "lwave": False,
            "nsw": 200,
            "sigma": 0.05,
        }
        flags = merge_dicts(defaults, swaps, remove_none=True)

        atoms = SmartVasp(atoms, preset=self.preset, **flags)
        atoms = run_calc(atoms)
        summary = summarize_run(atoms, additional_fields={"name": self.name})

        return summary



import warnings
from copy import deepcopy
from typing import Any, Dict, List

import numpy as np
from ase.atoms import Atoms
from ase.build import molecule
from ase.collections import g2
from pymatgen.analysis.adsorption import AdsorbateSiteFinder
from pymatgen.core import Structure
from pymatgen.core.surface import Slab, center_slab, generate_all_slabs
from pymatgen.io.ase import AseAtomsAdaptor

# NOTES:
# - Anytime an Atoms object is converted to a pmg structure, make sure
# to reattach any .info flags to the Atoms object, e.g. via `new_atoms.info = atoms.info.copy()``.
# Note that atoms.info is mutable, so copy it!
# - All major functions should take in Atoms by default and reutrn Atoms
# by default. Pymatgen structures can be returned with an optional kwarg.
# - If you modify the properties of an input Atoms object in any way, make sure to do so
# on a deepcopy because Atoms objects are mutable.


def flip_atoms(
    atoms: Atoms | Structure | Slab, return_struct: bool = False
) -> Atoms | Structure | Slab:
    """
    Convenience function for vertically flipping periodic atoms or structures

    Parameters
    ----------
    atoms
        .Atoms/structure to flip
    return_struct
        True if a Pymatgen structure object should be returned.
        False if an ASE atoms object should be returned

    Returns
    -------
    Atoms | Structure | Slab
        Inverted slab
    """

    if isinstance(atoms, Atoms):
        new_atoms = deepcopy(atoms)
        atoms_info = atoms.info.copy()
    else:
        new_atoms = AseAtomsAdaptor.get_atoms(atoms)
        atoms_info = {}

    new_atoms.rotate(180, "x")
    new_atoms.wrap()

    new_atoms.info = atoms_info
    if return_struct:
        new_atoms = AseAtomsAdaptor.get_structure(new_atoms)

    return new_atoms


def make_slabs_from_bulk(
    atoms: Atoms,
    max_index: int = 1,
    min_slab_size: float = 10.0,
    min_length_width: float = 8.0,
    min_vacuum_size: float = 20.0,
    z_fix: None | float = 2.0,
    flip_asymmetric: bool = True,
    allowed_surface_atoms: None | List[str] = None,
    **slabgen_kwargs,
) -> None | List[Atoms]:
    """
    Function to make slabs from a bulk atoms object.

    Parameters
    ----------
    atoms
        bulk atoms
    max_index
        Maximum Miller index for slab generation
    min_slab_size
        Minimum slab size (depth) in angstroms
    min_length_width
        Minimum length and width of the slab in angstroms
    min_vacuum_size
        Minimum vacuum size in angstroms
    z_fix
        Distance (in angstroms) from top of slab for which atoms should be fixed
    flip_asymmetric
        If an asymmetric surface should be flipped and added to the list
    allowed_surface_atoms
        List of chemical symbols that must be present on the surface of the slab otherwise the slab will be discarded, e.g. ["Cu", "Ni"]
    **slabgen_kwargs: keyword arguments to pass to the pymatgen generate_all_slabs() function

    Returns
    -------
    Optional[List[.Atoms]]
        All generated slabs
    """

    # Note: This will not work properly for 2D structures. See Oxana/Martin's code
    # for adjustments for 2D: https://github.com/oxana-a/atomate/blob/ads_wf/atomate/vasp/firetasks/adsorption_tasks.py

    # Use pymatgen to generate slabs
    struct = AseAtomsAdaptor.get_structure(atoms)
    atoms_info = atoms.info.copy()

    # Make all the slabs
    slabs = generate_all_slabs(
        struct,
        max_index,
        min_slab_size,
        min_vacuum_size,
        center_slab=True,
        **slabgen_kwargs,
    )

    # If the two terminations are not equivalent, make new slab
    # by inverting the original slab and add it to the list
    if flip_asymmetric:

        new_slabs = []
        for slab in slabs:
            if not slab.is_symmetric():

                # Flip the slab and its oriented unit cell
                new_slab = flip_atoms(slab, return_struct=True)
                new_oriented_unit_cell = flip_atoms(
                    slab.oriented_unit_cell, return_struct=True
                )

                # Reconstruct the full slab object, noting the new
                # shift and oriented unit cell
                new_slab = Slab(
                    new_slab.lattice,
                    new_slab.species,
                    coords=new_slab.frac_coords,
                    miller_index=slab.miller_index,
                    oriented_unit_cell=new_oriented_unit_cell,
                    shift=-slab.shift,
                    scale_factor=slab.scale_factor,
                    site_properties=new_slab.site_properties,
                )

                # It looks better to center the inverted slab so we do
                # that here.
                new_slab = center_slab(new_slab)

                # Add the new slab to the list
                new_slabs.append(new_slab)

        slabs.extend(new_slabs)

    # For each slab, make sure the lengths and widths are large enough
    # and fix atoms z_fix away from the top of the slab.
    slabs_with_props = []
    for slab in slabs:

        # Supercell creation (if necessary)
        a_factor = int(np.ceil(min_length_width / slab.lattice.abc[0]))
        b_factor = int(np.ceil(min_length_width / slab.lattice.abc[1]))
        slab_with_props = slab.copy()
        slab_with_props.make_supercell([a_factor, b_factor, 1])

        # Apply constraints by distance from top surface
        # This does not actually create an adsorbate. It is just a
        # useful function for finding surface vs. subsurface sites
        if z_fix:
            slab_with_props = AdsorbateSiteFinder(
                slab_with_props, selective_dynamics=True, height=z_fix
            ).slab

            surface_species = [
                site.specie.symbol
                for site in slab_with_props
                if site.properties["surface_properties"] == "surface"
            ]

            # Check that the desired atoms are on the surface
            if allowed_surface_atoms and not any(
                [
                    allowed_surface_atom in surface_species
                    for allowed_surface_atom in allowed_surface_atoms
                ]
            ):
                continue

        # Add slab to list
        slabs_with_props.append(slab_with_props)

    if len(slabs_with_props) == 0:
        return None

    # Make atoms objects and store slab stats
    final_slabs = []
    for slab_with_props in slabs_with_props:
        final_slab = AseAtomsAdaptor.get_atoms(slab_with_props)
        slab_stats = {
            "bulk": atoms,
            "miller_index": slab_with_props.miller_index,
            "shift": round(slab_with_props.shift, 3),
            "scale_factor": slab_with_props.scale_factor,
        }
        final_slab.info = atoms_info.copy()
        final_slab.info["slab_stats"] = slab_stats
        final_slabs.append(final_slab)

    return final_slabs


def make_max_slabs_from_bulk(
    atoms: Atoms,
    max_slabs: None | int = None,
    max_index: int = 1,
    min_slab_size: float = 10.0,
    min_length_width: float = 8.0,
    min_vacuum_size: float = 20.0,
    z_fix: float = 2.0,
    flip_asymmetric: bool = True,
    allowed_surface_atoms: bool = None,
    **slabgen_kwargs,
) -> List[Atoms]:

    """
    Generate no more than max_slabs number of slabs from a bulk structure.
    The procedure is as follows:
    1. Generate all slabs
    2. If number of slabs is greater than max_slabs, tune ftol from 0.1 to 0.8
    in increments of 0.1. This reduces the number of vertical shifts to consider.
    3. If number of slabs is still greater than max_slabs, only return the slabs
    with the fewest number of atoms per cell such that the returned amount is
    less than or equal to max_slabs.

    Parameters
    ----------
    atoms
        Bulk structure to generate slabs from
    max_slabs
        Maximum number of slabs to generate
    max_index
        Maximum Miller index for slab generation
    min_slab_size
        Minimum slab size (depth) in angstroms
    min_length_width
        Minimum length and width of the slab in angstroms
    min_vacuum_size
        Minimum vacuum size in angstroms
    z_fix
        Distance (in angstroms) from top of slab for which atoms should be fixed
    flip_asymmetric
        If an asymmetric surface should be flipped and added to the list
    allowed_surface_atoms
        List of chemical symbols that must be present on the surface of the slab otherwise
        the slab will be discarded, e.g. ["Cu", "Ni"]
    **slabgen_kwargs: keyword arguments to pass to the pymatgen generate_all_slabs() function

    Returns:
    --------
    List[.Atoms]
        List of slabs

    """

    slabs = make_slabs_from_bulk(
        atoms,
        max_index=max_index,
        min_slab_size=min_slab_size,
        min_length_width=min_length_width,
        min_vacuum_size=min_vacuum_size,
        z_fix=z_fix,
        flip_asymmetric=flip_asymmetric,
        allowed_surface_atoms=allowed_surface_atoms,
        **slabgen_kwargs,
    )

    # Try to reduce the number of slabs if the user really wants it...
    # (desperate times call for desperate measures)
    if max_slabs and slabs is not None and len(slabs) > max_slabs:

        if len(slabs) > max_slabs:
            warnings.warn(
                f"You requested {max_slabs} slabs, but {len(slabs)} were generated. Tuning ftol in generate_all_slabs() to try to reduce the number of slabs, at the expense of sampling fewer surface configurations.",
                UserWarning,
            )
            for ftol in np.arange(0.1, 0.9, 0.1):
                slabgen_kwargs["ftol"] = ftol
                slabs_ftol = make_slabs_from_bulk(
                    atoms,
                    max_index=max_index,
                    min_slab_size=min_slab_size,
                    min_length_width=min_length_width,
                    min_vacuum_size=min_vacuum_size,
                    z_fix=z_fix,
                    flip_asymmetric=flip_asymmetric,
                    allowed_surface_atoms=allowed_surface_atoms,
                    **slabgen_kwargs,
                )
                if len(slabs_ftol) < len(slabs):
                    slabs = slabs_ftol
                if len(slabs) <= max_slabs:
                    break

        if len(slabs) > max_slabs:
            warnings.warn(
                f"You requested {max_slabs} slabs, but {len(slabs)} were generated. Could not reduce further. Picking the smallest slabs by number of atoms.",
                UserWarning,
            )
            slabs.sort(key=lambda s: len(s))
            slabs = slabs[0:max_slabs]

    return slabs


# TODO: We need a method to orient adsorbate via a kwarg
def make_adsorbate_structures(
    atoms: Atoms,
    adsorbate: Atoms | str,
    min_distance: float = 2.0,
    modes: List[str] = ["ontop", "bridge", "hollow"],
    allowed_surface_symbols: None | List[str] = None,
    allowed_surface_indices: None | List[int] = None,
    ads_site_finder_kwargs: None | Dict[str, Any] = None,
    find_ads_sites_kwargs: None | Dict[str, Any] = None,
) -> None | List[Atoms]:
    """
    Add a single adsorbate to a structure for every requested adsorption mode

    Parameters
    ----------
    atoms
        The atoms to add adsorbates to.
    adsorbate
        The adsorbate to add. If a string, it will pull from ase.collections.g2
        Note: It will be placed on the surface in the exact input orientation provided by the user (the adsorption mode is
        along the c axis and the coordinating atom is the one in the -z direction).
    min_distance
        The distance between the adsorbate and the surface site.
    modes
        The adsorption mode(s) to consider. Options include: "ontop", "bridge", "hollow", "subsurface".
    allowed_surface_symbols
        The symbols of surface atoms to consider. If None, will use all surface atoms.
    allowed_surface_indices
        The indices of surface atoms to consider. If None, will use all surface atoms. Generally used if a specific site is to be excluded from the set.
    ads_site_finder_kwargs
        The keyword arguments to pass to the AdsorbateSiteFinder().
    find_ads_sites_kwargs
        The keyword arguments to pass to AdsorbateSiteFinder.find_adsorption_sites().

    Returns:
    --------
    Optional[List[ase.Atoms]]
        The structures with adsorbates

    """

    ads_site_finder_kwargs = ads_site_finder_kwargs or {}
    find_ads_sites_kwargs = find_ads_sites_kwargs or {}

    # Check for double-used parameters
    if min_distance and "distance" in find_ads_sites_kwargs:
        raise ValueError(
            "Cannot specify both min_distance and find_ads_sites_kwargs['distance']",
        )
    if modes and "positions" in find_ads_sites_kwargs:
        raise ValueError(
            "Cannot specify both modes and find_ads_sites_kwargs['positions']",
        )
    find_ads_sites_kwargs["distance"] = min_distance
    find_ads_sites_kwargs["positions"] = [mode.lower() for mode in modes]

    # Check the provided surface indices are reasonable
    atom_indices = [atom.index for atom in atoms]
    if allowed_surface_indices and not all(
        [idx in atom_indices for idx in allowed_surface_indices]
    ):
        raise ValueError(
            "All indices in allowed_surface_indices must be in atoms.",
            allowed_surface_indices,
            atom_indices,
        )

    if isinstance(adsorbate, str):
        # Get adsorbate if string
        if adsorbate in g2.names:
            adsorbate = molecule(adsorbate)
            # Remove any adsorbate magmoms from the g2 collection. I find
            # it very bothersome that ASE automatically adds magnetic moments
            # without the user's consent or knowledge. Leave that to the user
            # except for O2.
            if adsorbate.get_chemical_formula() != "O2":
                adsorbate.set_initial_magnetic_moments(None)
        else:
            raise ValueError(f"{adsorbate} is not in the G2 database.")

    # Add 0.0 initial magmoms to atoms/adsorbate if needed
    if atoms.has("initial_magmoms") and not adsorbate.has("initial_magmoms"):
        adsorbate.set_initial_magnetic_moments([0.0] * len(adsorbate))
    if adsorbate.has("initial_magmoms") and not atoms.has("initial_magmoms"):
        atoms.set_initial_magnetic_moments([0.0] * len(atoms))

    # Make a Pymatgen structure and molecule
    struct = AseAtomsAdaptor.get_structure(atoms)
    mol = AseAtomsAdaptor.get_molecule(adsorbate)

    # Get the adsorption sites
    ads_finder = AdsorbateSiteFinder(struct, **ads_site_finder_kwargs)
    ads_sites = ads_finder.find_adsorption_sites(**find_ads_sites_kwargs)

    # Find and add the adsorbates
    new_atoms = []
    for mode, ads_coords in ads_sites.items():

        # Check if mode is in desired list
        if mode not in modes:
            continue

        for ads_coord in ads_coords:

            # Place adsorbate
            struct_with_adsorbate = ads_finder.add_adsorbate(mol, ads_coord)

            # Convert back to Atoms object
            atoms_with_adsorbate = AseAtomsAdaptor.get_atoms(struct_with_adsorbate)

            # Get distance matrix between adsorbate binding atom and surface
            adsorbate_index = len(atoms) + np.argmin(atom.z for atom in adsorbate)
            d = atoms_with_adsorbate.get_all_distances(mic=True)
            d = d[atom_indices, :]
            d = d[:, adsorbate_index]

            # Find closest surface atoms
            min_d = min(d)
            surface_atom_indices = [
                i for i, val in enumerate(d) if (min_d - 0.01) <= val <= (min_d + 0.01)
            ]
            surface_atom_symbols = atoms_with_adsorbate[
                surface_atom_indices
            ].get_chemical_symbols()

            # Check if surface binding site is not in the specified
            # user list. If so, skip this one
            if allowed_surface_symbols:
                if not any(
                    [
                        surface_atom_symbol in allowed_surface_symbols
                        for surface_atom_symbol in surface_atom_symbols
                    ]
                ):
                    continue

            if allowed_surface_indices:
                if not any(
                    [
                        surface_atom_idx in allowed_surface_indices
                        for surface_atom_idx in surface_atom_indices
                    ]
                ):
                    continue

            # Store adsorbate info
            atoms_with_adsorbate.info = atoms.info.copy()
            ads_stats = {
                "adsorbate": adsorbate,
                "initial_mode": mode,
                "surface_atoms_symbols": surface_atom_symbols,
                "surface_atoms_indices": surface_atom_indices,
            }
            if atoms_with_adsorbate.info.get("adsorbates", None) is None:
                atoms_with_adsorbate.info["adsorbates"] = [ads_stats]
            else:
                atoms_with_adsorbate.info["adsorbates"].extend([ads_stats])

            # Add slab+adsorbate to list
            new_atoms.append(atoms_with_adsorbate)

    return new_atoms



import multiprocessing
from dataclasses import dataclass
from typing import Any, Dict

from ase.atoms import Atoms
from ase.calculators.orca import ORCA
from jobflow import Maker, job

from quacc.schemas.cclib import summarize_run
from quacc.util.basics import merge_dicts
from quacc.util.calc import run_calc


@dataclass
class StaticMaker(Maker):
    """
    Class to carry out a single-point calculation.

    Parameters
    ----------
    name
        Name of the job.
    xc
        Exchange-correlation functional
    basis
        Basis set
    input_swaps
        Dictionary of orcasimpleinput swaps for the calculator.
        To enable new entries, set the value as True.
        To remove entries from the defaults, set the value as None/False.
    block_swaps
        Dictionary of orcablocks swaps for the calculator.
        To enable new entries, set the value as True.
        To remove entries from the defaults, set the value as None/False.
    """

    name: str = "ORCA-Static"
    xc: str = "wb97x-d"
    basis: str = "def2-tzvp"
    input_swaps: Dict[str, Any] = None
    block_swaps: Dict[str, Any] = None

    @job
    def make(
        self, atoms: Atoms, charge: int = None, mult: int = None
    ) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object
        charge
            Charge of the system. If None, this is determined from the sum of
            atoms.get_initial_charges().
        mult
            Multiplicity of the system. If None, this is determined from 1+ the sum
            of atoms.get_initial_magnetic_moments().

        Returns
        -------
        Dict
            Summary of the run.
        """
        input_swaps = self.input_swaps or {}
        block_swaps = self.block_swaps or {}
        if not any([k for k in block_swaps if "nprocs" in k.lower()]):
            nprocs = multiprocessing.cpu_count()
            block_swaps[f"%pal nprocs {nprocs} end"] = True

        default_inputs = {
            self.xc: True,
            self.basis: True,
            "sp": True,
            "slowconv": True,
            "normalprint": True,
        }
        default_blocks = {}

        inputs = merge_dicts(
            default_inputs, input_swaps, remove_none=True, remove_false=True
        )
        blocks = merge_dicts(
            default_blocks, block_swaps, remove_none=True, remove_false=True
        )
        orcasimpleinput = " ".join(list(inputs.keys()))
        orcablocks = " ".join(list(blocks.keys()))

        atoms.calc = ORCA(
            orcasimpleinput=orcasimpleinput,
            orcablocks=orcablocks,
            charge=charge if charge else round(sum(atoms.get_initial_charges())),
            mult=mult if mult else round(1 + sum(atoms.get_initial_magnetic_moments())),
        )
        atoms = run_calc(atoms)
        summary = summarize_run(
            atoms, "orca.out", additional_fields={"name": self.name}
        )

        return summary


@dataclass
class RelaxMaker(Maker):
    """
    Class to carry out a geometry optimization.

    Parameters
    ----------
    name
        Name of the job.
    xc
        Exchange-correlation functional
    basis
        Basis set
    freq
        If a requency calculation should be carried out.
    input_swaps
        Dictionary of orcasimpleinput swaps for the calculator.
        To enable new entries, set the value as True.
        To remove entries from the defaults, set the value as None/False.
    block_swaps
        Dictionary of orcablocks swaps for the calculator.
        To enable new entries, set the value as True.
        To remove entries from the defaults, set the value as None/False.
    """

    name: str = "ORCA-Relax"
    xc: str = "wb97x-d"
    basis: str = "def2-tzvp"
    freq: bool = False
    input_swaps: Dict[str, Any] = None
    block_swaps: Dict[str, Any] = None

    @job
    def make(
        self, atoms: Atoms, charge: int = None, mult: int = None
    ) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object
        charge
            Charge of the system. If None, this is determined from the sum of
            atoms.get_initial_charges().
        mult
            Multiplicity of the system. If None, this is determined from 1+ the sum
            of atoms.get_initial_magnetic_moments().

        Returns
        -------
        Dict
            Summary of the run.
        """
        input_swaps = self.input_swaps or {}
        block_swaps = self.block_swaps or {}
        if not any([k for k in block_swaps if "nprocs" in k.lower()]):
            nprocs = multiprocessing.cpu_count()
            block_swaps[f"%pal nprocs {nprocs} end"] = True

        default_inputs = {
            self.xc: True,
            self.basis: True,
            "opt": True,
            "slowconv": True,
            "normalprint": True,
            "freq": True if self.freq else None,
        }
        default_blocks = {}

        inputs = merge_dicts(
            default_inputs, input_swaps, remove_none=True, remove_false=True
        )
        blocks = merge_dicts(
            default_blocks, block_swaps, remove_none=True, remove_false=True
        )
        orcasimpleinput = " ".join(list(inputs.keys()))
        orcablocks = " ".join(list(blocks.keys()))

        atoms.calc = ORCA(
            orcasimpleinput=orcasimpleinput,
            orcablocks=orcablocks,
            charge=charge if charge else round(sum(atoms.get_initial_charges())),
            mult=mult if mult else round(1 + sum(atoms.get_initial_magnetic_moments())),
        )
        atoms = run_calc(atoms)
        summary = summarize_run(
            atoms, "orca.out", additional_fields={"name": self.name}
        )

        return summary



from dataclasses import dataclass
from typing import Any, Dict

from ase.atoms import Atoms
from jobflow import Flow, Maker, Response, job

from quacc.calculators.vasp import SmartVasp
from quacc.schemas.vasp import summarize_run
from quacc.util.basics import merge_dicts
from quacc.util.calc import run_calc
from quacc.util.slabs import make_adsorbate_structures, make_max_slabs_from_bulk


@dataclass
class SlabRelaxMaker(Maker):
    """
    Class to relax a slab.

    Parameters
    ----------
    name
        Name of the job.
    preset
        Preset to use.
    swaps
        Dictionary of custom kwargs for the calculator.
    """

    name: str = "VASP-SlabRelax"
    preset: None | str = None
    swaps: Dict[str, Any] = None

    @job
    def make(self, atoms: Atoms) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object

        Returns
        -------
        Dict:
            Summary of the run.
        """
        swaps = self.swaps or {}
        defaults = {
            "auto_dipole": True,
            "ediff": 1e-5,
            "ediffg": -0.02,
            "isif": 2,
            "ibrion": 2,
            "ismear": 0,
            "isym": 0,
            "lcharg": False,
            "lwave": False,
            "nsw": 200,
            "sigma": 0.05,
        }
        flags = merge_dicts(defaults, swaps, remove_none=True)

        atoms = SmartVasp(atoms, preset=self.preset, **flags)
        atoms = run_calc(atoms)
        summary = summarize_run(atoms, additional_fields={"name": self.name})

        return summary


@dataclass
class SlabStaticMaker(Maker):
    """
    Class to carry out a single-point calculation on a slab.

    Parameters
    ----------
    name
        Name of the job.
    preset
        Preset to use.
    swaps
        Dictionary of custom kwargs for the calculator.
    """

    name: str = "VASP-SlabStatic"
    preset: None | str = None
    swaps: Dict[str, Any] = None

    @job
    def make(self, atoms: Atoms) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object

        Returns
        -------
        Dict
            Summary of the run.
        """
        swaps = self.swaps or {}
        defaults = {
            "auto_dipole": True,
            "ediff": 1e-6,
            "ismear": -5,
            "isym": 2,
            "laechg": True,
            "lcharg": True,
            "lvhar": True,
            "lwave": True,
            "nedos": 5001,
            "nsw": 0,
            "sigma": 0.05,
        }
        flags = merge_dicts(defaults, swaps, remove_none=True)

        atoms = SmartVasp(atoms, preset=self.preset, **flags)
        atoms = run_calc(atoms)
        summary = summarize_run(atoms, additional_fields={"name": self.name})

        return summary


@dataclass
class BulkToSlabMaker(Maker):
    """
    Class to convert a bulk structure to a slab,
    along with the relaxations and statics for the slabs.

    Parameters
    ----------
    name
        Name of the job.
    preset
        Preset to use. Applies to all jobs in the flow.
    slab_relax_maker
        Maker to use for the SlabRelax job.
    slab_static_maker
        Default to use for the SlabStatic job.
    swaps
        Dictionary of custom kwargs for the calculator.
        Applies to all jobs in the flow.
    """

    name: str = "VASP-BulkToSlab"
    preset: None | str = None
    slab_relax_maker: Maker = SlabRelaxMaker()
    slab_static_maker: Maker = SlabStaticMaker()
    swaps: Dict[str, Any] = None

    @job
    def make(
        self, atoms: Atoms, max_slabs: None | int = None, **slabgen_kwargs
    ) -> Response:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object
        max_slabs
            Maximum number of slabs to make
        slabgen_kwargs
            Additional keyword arguments to pass to make_max_slabs_from_bulk()

        Returns
        -------
        Response
            A Flow of relaxation and static jobs for the generated slabs.
        """
        slabgen_kwargs = slabgen_kwargs or {}
        self.slab_relax_maker.preset = self.slab_relax_maker.preset or self.preset
        self.slab_static_maker.preset = self.slab_static_maker.preset or self.preset
        self.slab_relax_maker.swaps = self.slab_relax_maker.swaps or self.swaps
        self.slab_static_maker.swaps = self.slab_static_maker.swaps or self.swaps

        slabs = make_max_slabs_from_bulk(atoms, max_slabs=max_slabs, **slabgen_kwargs)
        jobs = []
        outputs = []
        for slab in slabs:
            relax_job = self.slab_relax_maker.make(slab)
            jobs.append(relax_job)
            outputs.append(relax_job.output)

            static_job = self.slab_static_maker.make(relax_job.output["atoms"])
            jobs.append(static_job)
            outputs.append(static_job.output)

        return Response(replace=Flow(jobs))


@dataclass
class SlabToAdsSlabMaker(Maker):
    """
    Class to convert a slab structure to one with adsorbates present,
    along with the relaxations and statics for the slab-adsorbate systems.

    Parameters
    ----------
    name
        Name of the job.
    preset
        Preset to use. Applies to all jobs in the flow.
    slab_relax_maker
        Maker to use for the SlabRelax job.
    slab_static_maker
        Default to use for the SlabStatic job.
    swaps
        Dictionary of custom kwargs for the calculator.
        Applies to all jobs in the flow.
    """

    name: str = "VASP-SlabToAdsSlab"
    preset: None | str = None
    swaps: Dict[str, Any] = None
    slab_relax_maker: Maker = SlabRelaxMaker()
    slab_static_maker: Maker = SlabStaticMaker()

    @job
    def make(
        self, atoms: Atoms, adsorbate: Atoms | str, **slabgen_ads_kwargs
    ) -> Response:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object for the structure.
        adsorbate
            .Atoms object for the adsorbate.
        slabgen_ads_kwargs
            Additional keyword arguments to pass to make_adsorbate_structures()

        Returns
        -------
        Response
            A Flow of relaxation and static jobs for the generated slabs with adsorbates.
        """
        slabgen_ads_kwargs = slabgen_ads_kwargs or {}
        self.slab_relax_maker.preset = self.slab_relax_maker.preset or self.preset
        self.slab_static_maker.preset = self.slab_static_maker.preset or self.preset
        self.slab_relax_maker.swaps = self.slab_relax_maker.swaps or self.swaps
        self.slab_static_maker.swaps = self.slab_static_maker.swaps or self.swaps

        slabs = make_adsorbate_structures(atoms, adsorbate, **slabgen_ads_kwargs)

        jobs = []
        outputs = []
        for slab in slabs:
            relax_job = self.slab_relax_maker.make(slab)
            jobs.append(relax_job)
            outputs.append(relax_job.output)

            static_job = self.slab_static_maker.make(relax_job.output["atoms"])
            jobs.append(static_job)
            outputs.append(static_job.output)

        return Response(replace=Flow(jobs))



from ase.atoms import Atoms
from ase.io.jsonio import decode, encode

from quacc._version import __version__


def atoms_as_dict(self):
    # Uses Monty's MSONable spec
    # Normally, we would want to this to be a wrapper around atoms.todict() with @module and
    # @class key-value pairs inserted. However, atoms.todict()/atoms.fromdict() does not currently
    # work properly with constraints.

    d = {"@module": "ase.atoms", "@class": "Atoms", "atoms_json": encode(self)}

    return d


def atoms_from_dict(d):
    # Uses Monty's MSONable spec
    # Normally, we would want to have this be a wrapper around atoms.fromdict()
    # that just ignores the @module/@class key-value pairs. However, atoms.todict()/atoms.fromdict()
    # does not currently work properly with constraints.

    atoms = decode(d["atoms_json"])

    return atoms


Atoms.as_dict = atoms_as_dict
Atoms.from_dict = atoms_from_dict



from ase.build import bulk
from monty.json import MontyDecoder, jsanitize

from quacc.schemas.atoms import atoms_to_metadata


def test_atoms_to_metadata():

    atoms = bulk("Cu")
    atoms.info["test"] = "hi"
    results = atoms_to_metadata(atoms)
    assert results["atoms"].info.get("test", None) == "hi"

    atoms = bulk("Cu")
    atoms.info["test"] = "hi"
    results = atoms_to_metadata(atoms, strip_info=True)
    assert results["atoms"] == atoms
    assert results["atoms"].info == {}
    assert results["nsites"] == len(atoms)
    assert results["atoms_info"].get("test", None) == "hi"

    atoms = bulk("Cu")
    results = atoms_to_metadata(atoms, get_metadata=False)
    assert results["atoms"] == atoms
    assert results.get("nsites", None) is None

    atoms = bulk("Cu")
    parent = bulk("Al") * (2, 1, 1)
    atoms.info = {"parent": parent}
    results = atoms_to_metadata(atoms)
    assert results["atoms_info"]["parent"]["atoms"] == parent
    assert results["atoms_info"]["parent"]["nsites"] == len(parent)

    # test document can be jsanitized and decoded
    d = jsanitize(results, strict=True, enum_values=True)
    MontyDecoder().process_decoded(d)



from dataclasses import dataclass
from typing import Any, Dict

from ase.atoms import Atoms
from ase.calculators.emt import EMT
from ase.optimize import FIRE
from ase.optimize.optimize import Optimizer
from jobflow import Maker, job

from quacc.schemas.calc import summarize_run

# NOTE: This set of recipes is mainly for demonstration purposes


@dataclass
class StaticMaker(Maker):
    """
    Class to carry out a single-point calculation.

    Parameters
    ----------
    name
        Name of the job.
    asap_cutoff
        If an ASAP-style cutoff should be used.
    """

    name: str = "EMT-Static"
    asap_cutoff: bool = False

    @job
    def make(self, atoms: Atoms) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object`

        Returns
        -------
        Dict
            Summary of the run.
        """
        atoms.calc = EMT(asap_cutoff=self.asap_cutoff)
        atoms.get_potential_energy()
        summary = summarize_run(atoms, additional_fields={"name": self.name})

        return summary


@dataclass
class RelaxMaker(Maker):
    """
    Class to carry out a geometry optimization.

    Parameters
    ----------
    name
        Name of the job.
    asap_cutoff
        If an ASAP-style cutoff should be used.
    optimizer
        .Optimizer class to use for the relaxation.
    fmax
        Tolerance for the force convergence (in eV/A).
    opt_kwargs
        Dictionary of kwargs for the optimizer.
    """

    name: str = "EMT-Relax"
    asap_cutoff: bool = False
    optimizer: Optimizer = FIRE
    fmax: float = 0.03
    opt_kwargs: Dict[str, Any] = None

    @job
    def make(self, atoms: Atoms) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object

        Returns
        -------
        Dict
            Summary of the run.
        """
        opt_kwargs = self.opt_kwargs or {}

        # We always want to save the logfile and trajectory, so we will set some default
        # values if not specified by the user (and then remove them from the **opt_kwargs)
        logfile = opt_kwargs.get("logfile", None) or "opt.log"
        trajectory = opt_kwargs.get("trajectory", None) or "opt.traj"
        opt_kwargs.pop("logfile", None)
        opt_kwargs.pop("trajectory", None)

        atoms.calc = EMT(asap_cutoff=self.asap_cutoff)
        dyn = self.optimizer(
            atoms, logfile=logfile, trajectory=trajectory, **opt_kwargs
        )
        dyn.run(fmax=self.fmax)
        summary = summarize_run(atoms, additional_fields={"name": self.name})

        return summary



import multiprocessing
from dataclasses import dataclass
from typing import Any, Dict

from ase.atoms import Atoms
from ase.calculators.gaussian import Gaussian
from jobflow import Maker, job

from quacc.schemas.cclib import summarize_run
from quacc.util.basics import merge_dicts
from quacc.util.calc import run_calc


@dataclass
class StaticMaker(Maker):
    """
    Class to carry out a single-point calculation.

    Parameters
    ----------
    name
        Name of the job.
    xc
        Exchange-correlation functional
    basis
        Basis set
    pop
        Type of population analysis to perform, if any
    molden
        Whether to write a molden file for orbital visualization
    swaps
        Dictionary of custom kwargs for the calculator.
    """

    name: str = "Gaussian-Static"
    xc: str = "wb97x-d"
    basis: str = "def2-tzvp"
    pop: str = "hirshfeld"
    molden: bool = True
    swaps: Dict[str, Any] = None

    @job
    def make(
        self, atoms: Atoms, charge: int = None, mult: int = None
    ) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object
        charge
            Charge of the system. If None, this is determined from the sum of
            atoms.get_initial_charges().
        mult
            Multiplicity of the system. If None, this is determined from 1+ the sum
            of atoms.get_initial_magnetic_moments().

        Returns
        -------
        Dict
            Summary of the run.
        """
        swaps = self.swaps or {}
        defaults = {
            "mem": "16GB",
            "chk": "Gaussian.chk",
            "nprocshared": multiprocessing.cpu_count(),
            "xc": self.xc,
            "basis": self.basis,
            "charge": charge,
            "mult": mult,
            "sp": "",
            "scf": ["maxcycle=250", "xqc"],
            "integral": "ultrafine",
            "nosymmetry": "",
            "pop": self.pop,
            "gfinput": "" if self.molden else None,
            "ioplist": ["6/7=3"] if self.molden else None,
        }
        flags = merge_dicts(defaults, swaps, remove_none=True)

        atoms.calc = Gaussian(**flags)
        atoms = run_calc(atoms)
        summary = summarize_run(
            atoms, "Gaussian.log", additional_fields={"name": self.name}
        )

        return summary


@dataclass
class RelaxMaker(Maker):
    """
    Class to carry out a geometry optimization.

    Parameters
    ----------
    name
        Name of the job.
    xc
        Exchange-correlation functional
    basis
        Basis set
    freq
        If a requency calculation should be carried out.
    swaps
        Dictionary of custom kwargs for the calculator.
    """

    name: str = "Gaussian-Relax"
    xc: str = "wb97x-d"
    basis: str = "def2-tzvp"
    freq: bool = False
    swaps: Dict[str, Any] = None

    @job
    def make(
        self, atoms: Atoms, charge: int = None, mult: int = None
    ) -> Dict[str, Any]:
        """
        Make the run.

        Parameters
        ----------
        atoms
            .Atoms object
        charge
            Charge of the system. If None, this is determined from the sum of
            atoms.get_initial_charges().
        mult
            Multiplicity of the system. If None, this is determined from 1+ the sum
            of atoms.get_initial_magnetic_moments().

        Returns
        -------
        Dict
            Summary of the run.
        """
        swaps = self.swaps or {}
        defaults = {
            "mem": "16GB",
            "chk": "Gaussian.chk",
            "nprocshared": multiprocessing.cpu_count(),
            "xc": self.xc,
            "basis": self.basis,
            "charge": charge,
            "mult": mult,
            "opt": "",
            "scf": ["maxcycle=250", "xqc"],
            "integral": "ultrafine",
            "nosymmetry": "",
            "freq": "" if self.freq else None,
        }
        flags = merge_dicts(defaults, swaps, remove_none=True)

        atoms.calc = Gaussian(**flags)
        atoms = run_calc(atoms)
        summary = summarize_run(
            atoms, "Gaussian.log", additional_fields={"name": self.name}
        )

        return summary



import os
from pathlib import Path
from shutil import which

import pytest
from ase.io import read
from monty.json import MontyDecoder, jsanitize

from quacc.calculators.vasp import SmartVasp
from quacc.schemas.vasp import summarize_run

FILE_DIR = Path(__file__).resolve().parent

run1 = os.path.join(FILE_DIR, "vasp_run1")


def mock_bader_analysis(*args, **kwargs):
    # NOTE: This is hard-coded for vasp_run1 with its 16 atoms
    bader_stats = {
        "min_dist": [1.0] * 16,
        "atomic_volume": [1.0] * 16,
        "vacuum_charge": 1.0,
        "vacuum_volume": 1.0,
        "bader_version": 1.0,
        "reference_used": [0.0] * 16,
        "partial_charges": [-1.0] * 16,
        "spin_moments": [0.0] * 16,
    }
    return bader_stats


@pytest.fixture(autouse=True)
def patch_pop_analyses(monkeypatch):
    # Monkeypatch the Bader analysis
    monkeypatch.setattr(
        "quacc.schemas.vasp.run_bader",
        mock_bader_analysis,
    )


def test_summarize_run():

    # Make sure metadata is made
    atoms = read(os.path.join(run1, "OUTCAR.gz"))
    results = summarize_run(atoms, dir_path=run1)
    assert results["nsites"] == len(atoms)
    assert results["atoms"] == atoms
    assert results["output"]["energy"] == -33.15807349
    assert results.get("calcs_reversed", None) is None

    # Make sure default dir works
    cwd = os.getcwd()
    os.chdir(run1)
    summarize_run(atoms)
    os.chdir(cwd)

    # Make sure metadata is made
    atoms = read(os.path.join(run1, "OUTCAR.gz"))
    results = summarize_run(atoms, dir_path=run1, compact=False, remove_empties=True)
    assert results.get("calcs_reversed", None) is not None
    assert "author" not in results
    assert "additional_json" not in results
    assert "handler" not in results["custodian"][0]
    assert "corrections" not in results["custodian"][0]

    # Make sure null are not removed
    atoms = read(os.path.join(run1, "OUTCAR.gz"))
    results = summarize_run(atoms, dir_path=run1, remove_empties=False)
    assert results["author"] == None
    assert results["additional_json"] == {}
    assert results["custodian"][0]["handler"] is None
    assert results["custodian"][0]["corrections"] == []

    # Make sure info tags are handled appropriately
    atoms = read(os.path.join(run1, "CONTCAR.gz"))
    atoms.info["test_dict"] = {"hi": "there", "foo": "bar"}
    results = summarize_run(atoms, dir_path=run1)
    results_atoms = results["atoms"]
    assert atoms.info.get("test_dict", None) == {"hi": "there", "foo": "bar"}
    assert results.get("atoms_info", {}) != {}
    assert results["atoms_info"].get("test_dict", None) == {"hi": "there", "foo": "bar"}
    assert results_atoms.info.get("test_dict", None) == {"hi": "there", "foo": "bar"}

    # Make sure magnetic moments are handled appropriately
    atoms = read(os.path.join(run1, "CONTCAR.gz"))
    atoms.set_initial_magnetic_moments([3.14] * len(atoms))
    atoms = SmartVasp(atoms)
    atoms.calc.results = {"energy": -1.0, "magmoms": [2.0] * len(atoms)}
    results = summarize_run(atoms, dir_path=run1)
    results_atoms = results["atoms"]

    assert atoms.calc is not None
    assert atoms.get_initial_magnetic_moments().tolist() == [3.14] * len(atoms)

    assert results_atoms.get_initial_magnetic_moments().tolist() == [2.0] * len(atoms)
    assert results_atoms.calc is None

    # Make sure Atoms magmoms were not moved if specified
    atoms = read(os.path.join(run1, "CONTCAR.gz"))
    atoms.set_initial_magnetic_moments([3.14] * len(atoms))
    results = summarize_run(atoms, dir_path=run1, prep_next_run=False)
    assert atoms.get_initial_magnetic_moments().tolist() == [3.14] * len(atoms)
    results_atoms = results["atoms"]
    assert results_atoms.get_initial_magnetic_moments().tolist() == [3.14] * len(atoms)

    # test document can be jsanitized and decoded
    d = jsanitize(results, strict=True, enum_values=True)
    MontyDecoder().process_decoded(d)


def test_summarize_bader_run():
    # Make sure Bader works
    atoms = read(os.path.join(run1, "OUTCAR.gz"))
    results = summarize_run(atoms, dir_path=run1, bader=True)
    struct = results["output"]["structure"]
    assert struct.site_properties["bader_charge"] == [-1.0] * len(atoms)
    assert struct.site_properties["bader_spin"] == [0.0] * len(atoms)



import os
from copy import deepcopy
from pathlib import Path

import numpy as np
import pytest
from ase.build import bulk
from ase.calculators.singlepoint import SinglePointDFTCalculator
from ase.calculators.vasp import Vasp
from ase.constraints import FixAtoms, FixBondLength
from ase.io import read

from quacc.calculators.vasp import SmartVasp
from quacc.defaults.calcs import vasp as v
from quacc.util.atoms import prep_next_run

FILE_DIR = Path(__file__).resolve().parent
DEFAULT_CALCS_DIR = os.path.dirname(v.__file__)
ATOMS_MAG = read(os.path.join(FILE_DIR, "OUTCAR_mag.gz"))
ATOMS_NOMAG = read(os.path.join(FILE_DIR, "OUTCAR_nomag.gz"))
ATOMS_NOSPIN = read(os.path.join(FILE_DIR, "OUTCAR_nospin.gz"))


def test_vanilla_smartvasp():
    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, incar_copilot=False)
    assert atoms.calc.asdict() == Vasp().asdict()

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, custodian=False, incar_copilot=False)
    assert atoms.calc.asdict() == Vasp().asdict()

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, encut=None, incar_copilot=False)
    assert atoms.calc.asdict() == Vasp().asdict()


def test_presets():
    atoms = bulk("Co") * (2, 2, 1)
    atoms[-1].symbol = "Fe"

    atoms_fullpath = SmartVasp(
        atoms, preset=os.path.join(DEFAULT_CALCS_DIR, "BulkRelaxSet")
    )
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms == atoms_fullpath
    assert atoms.calc.xc.lower() == "pbe"
    assert atoms.calc.string_params["algo"] == "fast"
    assert atoms.calc.exp_params["ediff"] == 1e-5
    assert atoms.calc.int_params["isif"] == 3

    atoms = SmartVasp(atoms, xc="rpbe", preset="SlabRelaxSet")
    assert atoms.calc.xc.lower() == "rpbe"
    assert atoms.calc.string_params["algo"] == "fast"
    assert atoms.calc.exp_params["ediff"] == 1e-5
    assert atoms.calc.int_params["isif"] == 2

    atoms = SmartVasp(atoms, xc="scan", preset="MPScanRelaxSet")
    assert atoms.calc.xc.lower() == "scan"
    assert atoms.calc.string_params["algo"] == "all"
    assert atoms.calc.exp_params["ediff"] == 1e-5
    assert atoms.calc.int_params["isif"] == 3


def test_lmaxmix():
    atoms = bulk("Cu")
    atoms = SmartVasp(atoms)
    assert atoms.calc.int_params["lmaxmix"] == 4

    atoms = bulk("Ce")
    atoms = SmartVasp(atoms)
    assert atoms.calc.int_params["lmaxmix"] == 6

    atoms = bulk("Cu") * (2, 2, 2)
    atoms[-1].symbol = "Ce"
    atoms = SmartVasp(atoms)
    assert atoms.calc.int_params["lmaxmix"] == 6


def test_autodipole():
    atoms = bulk("Cu")
    com = atoms.get_center_of_mass(scaled=True)
    atoms = SmartVasp(atoms, auto_dipole=True)
    assert atoms.calc.bool_params["ldipol"] is True
    assert atoms.calc.int_params["idipol"] == 3
    assert np.array_equal(atoms.calc.list_float_params["dipol"], com)

    atoms = SmartVasp(atoms, auto_dipole=True, idipol=2)
    assert atoms.calc.bool_params["ldipol"] is True
    assert atoms.calc.int_params["idipol"] == 2
    assert np.array_equal(atoms.calc.list_float_params["dipol"], com)

    atoms = SmartVasp(atoms, preset="SlabRelaxSet")
    assert atoms.calc.bool_params["ldipol"] is True
    assert atoms.calc.int_params["idipol"] == 3
    assert np.array_equal(atoms.calc.list_float_params["dipol"], com)

    atoms = SmartVasp(atoms, preset="SlabRelaxSet", idipol=2)
    assert atoms.calc.bool_params["ldipol"] is True
    assert atoms.calc.int_params["idipol"] == 2
    assert np.array_equal(atoms.calc.list_float_params["dipol"], com)

    atoms = SmartVasp(atoms, auto_dipole=False, preset="SlabRelaxSet", idipol=2)
    assert atoms.calc.bool_params["ldipol"] is None
    assert atoms.calc.int_params["idipol"] == 2
    assert atoms.calc.list_float_params["dipol"] is None


def test_kspacing():
    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, kspacing=0.1, ismear=-5)
    assert atoms.calc.int_params["ismear"] == -5

    atoms = SmartVasp(atoms, kspacing=100, ismear=-5)
    assert atoms.calc.int_params["ismear"] == 0


def test_magmoms():
    atoms = bulk("Mg")
    atoms = SmartVasp(atoms)
    assert atoms.has("initial_magmoms") is False

    atoms = bulk("Mg")
    atoms.set_initial_magnetic_moments([3.14] * len(atoms))
    atoms = SmartVasp(atoms)
    assert atoms.get_initial_magnetic_moments().tolist() == [3.14] * len(atoms)

    atoms = bulk("Cu") * (2, 2, 1)
    atoms[-1].symbol = "Fe"
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.get_initial_magnetic_moments().tolist() == [2.0] * (len(atoms) - 1) + [
        5.0
    ]

    atoms = bulk("Zn") * (2, 2, 1)
    atoms[-1].symbol = "Fe"
    atoms = SmartVasp(atoms, preset="BulkRelaxSet", mag_default=2.5)
    assert atoms.get_initial_magnetic_moments().tolist() == [2.5] * (len(atoms) - 1) + [
        5.0
    ]

    atoms = bulk("Eu") * (2, 2, 1)
    atoms[-1].symbol = "Fe"
    atoms = SmartVasp(atoms, preset="SlabRelaxSet")
    assert atoms.get_initial_magnetic_moments().tolist() == [7.0] * (len(atoms) - 1) + [
        5.0
    ]

    atoms = bulk("Cu") * (2, 2, 1)
    atoms[-1].symbol = "Fe"
    atoms = SmartVasp(atoms, preset="MPScanRelaxSet")
    assert atoms.get_initial_magnetic_moments().tolist() == [1.0] * (len(atoms) - 1) + [
        5.0
    ]

    atoms = bulk("Cu") * (2, 2, 1)
    atoms[-1].symbol = "Fe"
    atoms.set_initial_magnetic_moments([3.14] * (len(atoms) - 1) + [1.0])
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.get_initial_magnetic_moments().tolist() == [3.14] * (
        len(atoms) - 1
    ) + [1.0]

    atoms = bulk("Co") * (2, 2, 1)
    atoms[-1].symbol = "Fe"
    atoms.set_initial_magnetic_moments([0.0] * len(atoms))
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.has("initial_magmoms") is True
    assert np.all(atoms.get_initial_magnetic_moments() == 0)

    atoms = deepcopy(ATOMS_MAG)
    mags = atoms.get_magnetic_moments()
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.array_equal(atoms.get_initial_magnetic_moments(), mags) is True

    atoms = deepcopy(ATOMS_MAG)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet", mag_cutoff=2.0)
    assert atoms.has("initial_magmoms") is True
    assert np.all(atoms.get_initial_magnetic_moments() == 0)

    atoms = deepcopy(ATOMS_MAG)
    assert atoms.get_magnetic_moments()[0] == 0.468
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.get_initial_magnetic_moments()[0] == 0.468

    atoms = deepcopy(ATOMS_NOMAG)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.has("initial_magmoms") is True
    assert np.all(atoms.get_initial_magnetic_moments() == 0)
    atoms.calc.results = {"energy": -1.0}  # mock calculation run
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.has("initial_magmoms") is True
    assert np.all(atoms.get_initial_magnetic_moments() == 0)

    atoms = deepcopy(ATOMS_NOMAG)
    calc = SinglePointDFTCalculator(atoms, **{"magmoms": [0.0] * len(atoms)})
    atoms.calc = calc
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == 0)

    atoms = deepcopy(ATOMS_NOMAG)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == 0)
    atoms.calc.results = {"magmoms": [0.0] * len(atoms)}
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == 0)

    atoms = deepcopy(ATOMS_NOMAG)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.has("initial_magmoms") is True
    assert np.all(atoms.get_initial_magnetic_moments() == 0)
    atoms.calc.results = {"magmoms": [1.0] * len(atoms)}
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == 1)

    atoms = deepcopy(ATOMS_NOSPIN)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.has("initial_magmoms") is True
    assert np.all(atoms.get_initial_magnetic_moments() == 0)
    atoms.calc.results = {"magmoms": [1.0] * len(atoms)}
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == 1)

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, preset="BulkRelaxSet", copy_magmoms=False)
    assert np.all(atoms.get_initial_magnetic_moments() == 2.0)
    atoms.calc.results = {"magmoms": [3.0] * len(atoms)}
    atoms = SmartVasp(atoms, preset="BulkRelaxSet", copy_magmoms=False)
    assert np.all(atoms.get_initial_magnetic_moments() == 2.0)

    atoms = bulk("Mg")
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == 1.0)
    atoms.calc.results = {"magmoms": [0.0] * len(atoms)}
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == 0)

    atoms = bulk("Mg")
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == 1.0)
    atoms.calc.results = {"magmoms": [-0.01] * len(atoms)}
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == 0)

    atoms = bulk("Mg")
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == 1.0)
    atoms.calc.results = {"magmoms": [-5] * len(atoms)}
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert np.all(atoms.get_initial_magnetic_moments() == -5)

    atoms = deepcopy(ATOMS_MAG)
    mags = atoms.get_magnetic_moments()
    atoms = prep_next_run(atoms)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.has("initial_magmoms") is True
    assert atoms.get_initial_magnetic_moments().tolist() == mags.tolist()

    atoms = deepcopy(ATOMS_NOMAG)
    atoms = prep_next_run(atoms)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.has("initial_magmoms") is True
    assert np.all(atoms.get_initial_magnetic_moments() == 0)

    atoms = deepcopy(ATOMS_NOSPIN)
    atoms = prep_next_run(atoms)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.has("initial_magmoms") is True
    assert np.all(atoms.get_initial_magnetic_moments() == 0)

    atoms = deepcopy(ATOMS_MAG)
    atoms = prep_next_run(atoms)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet", mag_cutoff=10.0)
    assert atoms.has("initial_magmoms") is True
    assert np.all(atoms.get_initial_magnetic_moments() == 0)

    atoms = bulk("Mg")
    atoms = SmartVasp(atoms)
    atoms.calc.results = {"energy": -1.0, "magmoms": [0.0] * len(atoms)}
    atoms = prep_next_run(atoms)
    atoms *= (2, 2, 2)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    np.all(atoms.get_initial_magnetic_moments() == 0.0)

    atoms = bulk("Mg")
    atoms = SmartVasp(atoms)
    atoms.calc.results = {"energy": -1.0, "magmoms": [-0.02] * len(atoms)}
    atoms = prep_next_run(atoms)
    atoms *= (2, 2, 2)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    np.all(atoms.get_initial_magnetic_moments() == 0.0)

    atoms = bulk("Mg")
    atoms = SmartVasp(atoms)
    atoms.calc.results = {"energy": -1.0}
    atoms = prep_next_run(atoms)
    atoms *= (2, 2, 2)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    np.all(atoms.get_initial_magnetic_moments() == 0.0)

    atoms = bulk("Mg")
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    atoms.calc.results = {"energy": -1.0, "magmoms": [3.14] * len(atoms)}
    atoms = prep_next_run(atoms)
    atoms *= (2, 2, 2)
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    np.all(atoms.get_initial_magnetic_moments() == 3.14)


def test_unused_flags():
    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, preset="BulkRelaxSet", potim=1.5, nsw=0)
    assert atoms.calc.int_params["nsw"] == 0
    assert atoms.calc.exp_params["ediffg"] is None
    assert atoms.calc.int_params["isif"] is None
    assert atoms.calc.float_params["potim"] is None

    atoms = SmartVasp(atoms, ldau=False, ldauprint=2)
    assert atoms.calc.int_params["ldauprint"] is None
    assert atoms.calc.bool_params["ldau"] is None


def test_lasph():
    atoms = bulk("Cu")

    atoms = SmartVasp(atoms, xc="rpbe")
    assert atoms.calc.bool_params["lasph"] is None

    atoms = SmartVasp(atoms, xc="m06l")
    assert atoms.calc.bool_params["lasph"] is True

    atoms = SmartVasp(atoms, xc="m06l", lasph=False)
    assert atoms.calc.bool_params["lasph"] is True

    atoms = SmartVasp(atoms, xc="hse06")
    assert atoms.calc.bool_params["lasph"] is True

    atoms = SmartVasp(atoms, xc="beef-vdw")
    assert atoms.calc.bool_params["lasph"] is True

    atoms = SmartVasp(atoms, ldau_luj={"Cu": {"L": 2, "U": 5, "J": 0.0}})
    assert atoms.calc.bool_params["lasph"] is True


def test_lmaxtau():
    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, lasph=True)
    assert atoms.calc.int_params["lmaxtau"] is None

    atoms = bulk("Ce")
    atoms = SmartVasp(atoms, lasph=True)
    assert atoms.calc.int_params["lmaxtau"] == 8

    atoms = bulk("Cu") * (2, 2, 2)
    atoms[-1].symbol = "Ce"
    atoms = SmartVasp(atoms, lasph=True)
    assert atoms.calc.int_params["lmaxtau"] == 8


def test_algo():
    atoms = bulk("Cu")

    atoms = SmartVasp(atoms, xc="rpbe")
    assert atoms.calc.string_params["algo"] is None

    atoms = SmartVasp(atoms, xc="m06l")
    assert atoms.calc.string_params["algo"] == "all"

    atoms = SmartVasp(atoms, xc="m06l", algo="fast")
    assert atoms.calc.string_params["algo"] == "all"

    atoms = SmartVasp(atoms, xc="hse06")
    assert atoms.calc.string_params["algo"] == "damped"
    assert atoms.calc.float_params["time"] == 0.5

    atoms[0].symbol = "H"
    atoms = SmartVasp(atoms, xc="hse06")
    assert atoms.calc.string_params["algo"] == "all"


def test_kpar():
    atoms = bulk("Cu")

    atoms = SmartVasp(atoms, kpts=[2, 2, 1], kpar=4)
    assert atoms.calc.int_params["kpar"] == 4

    atoms = SmartVasp(atoms, kpar=4)
    assert atoms.calc.int_params["kpar"] == 1


def test_isym():
    atoms = bulk("Cu")

    atoms = SmartVasp(atoms, isym=2)
    assert atoms.calc.int_params["isym"] == 2

    atoms = SmartVasp(atoms, isym=0)
    assert atoms.calc.int_params["isym"] == 0

    atoms = SmartVasp(atoms, xc="hse06", isym=2)
    assert atoms.calc.int_params["isym"] == 3

    atoms = SmartVasp(atoms, isym=2, nsw=100)
    assert atoms.calc.int_params["isym"] == 0

    atoms = SmartVasp(atoms, xc="hse06", isym=2, nsw=100)
    assert atoms.calc.int_params["isym"] == 0


def test_ncore():
    atoms = bulk("Cu")

    atoms = SmartVasp(atoms, ncore=16)
    assert atoms.calc.int_params["ncore"] == 1

    atoms = SmartVasp(atoms, npar=16)
    assert atoms.calc.int_params["ncore"] == 1
    assert atoms.calc.int_params["npar"] is None

    atoms *= (2, 2, 2)
    atoms = SmartVasp(atoms, ncore=4)
    assert atoms.calc.int_params["ncore"] == 4

    atoms = SmartVasp(atoms, ncore=4, lhfcalc=True)
    assert atoms.calc.int_params["ncore"] == 1

    atoms = SmartVasp(atoms, npar=4, lhfcalc=True)
    assert atoms.calc.int_params["ncore"] == 1
    assert atoms.calc.int_params["npar"] is None


def test_ismear():
    atoms = bulk("Cu")

    atoms = SmartVasp(atoms, nsw=10)
    assert atoms.calc.int_params["ismear"] is None

    atoms = SmartVasp(atoms, ismear=-5, nsw=10)
    assert atoms.calc.int_params["ismear"] == 1
    assert atoms.calc.float_params["sigma"] == 0.1

    atoms = SmartVasp(atoms, ismear=-5, nsw=0)
    assert atoms.calc.int_params["ismear"] == 0

    atoms = SmartVasp(atoms, kpts=(10, 10, 10), ismear=-5, nsw=0)
    assert atoms.calc.int_params["ismear"] == -5

    atoms = SmartVasp(atoms, ismear=0, nsw=10)
    assert atoms.calc.int_params["ismear"] == 0

    atoms = SmartVasp(atoms, nedos=3001, nsw=0)
    assert atoms.calc.int_params["ismear"] == 0

    atoms = SmartVasp(atoms, ismear=-5, nedos=3001, nsw=0)
    assert atoms.calc.int_params["ismear"] == 0

    atoms = SmartVasp(atoms, kpts=(10, 10, 10), nedos=3001, nsw=0)
    assert atoms.calc.int_params["ismear"] == -5

    atoms = SmartVasp(atoms, auto_kpts={"line_density": 100}, ismear=1)
    assert atoms.calc.int_params["ismear"] == 0
    assert atoms.calc.float_params["sigma"] == 0.01

    atoms = SmartVasp(atoms, auto_kpts={"line_density": 100}, ismear=0, sigma=1e-3)
    assert atoms.calc.int_params["ismear"] == 0
    assert atoms.calc.float_params["sigma"] == 1e-3

    atoms = SmartVasp(atoms, auto_kpts={"line_density": 100}, ismear=-5)
    assert atoms.calc.int_params["ismear"] == 0

    atoms = SmartVasp(atoms, kspacing=1.0, ismear=-5)
    assert atoms.calc.int_params["ismear"] == 0
    assert atoms.calc.float_params["sigma"] == 0.05

    atoms = SmartVasp(atoms, nsw=0, kspacing=1.0, ismear=1, sigma=0.1)
    assert atoms.calc.int_params["ismear"] == 1
    assert atoms.calc.float_params["sigma"] == 0.1

    atoms[0].symbol = "H"
    atoms = SmartVasp(atoms, kpts=(10, 10, 10), ismear=-5, nsw=10)
    assert atoms.calc.int_params["ismear"] == -5


def test_laechg():
    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, nsw=10, laechg=True)
    assert atoms.calc.bool_params["laechg"] is None

    atoms = SmartVasp(atoms, laechg=True)
    assert atoms.calc.bool_params["laechg"] is True

    atoms = SmartVasp(atoms, nsw=0, laechg=True)
    assert atoms.calc.bool_params["laechg"] is True

    atoms = SmartVasp(atoms, nsw=0, laechg=False)
    assert atoms.calc.bool_params["laechg"] is False


def test_ldauprint():
    atoms = bulk("Cu")

    atoms = SmartVasp(atoms, ldau=True)
    assert atoms.calc.int_params["ldauprint"] == 1

    atoms = SmartVasp(atoms, ldau=True, ldauprint=0)
    assert atoms.calc.int_params["ldauprint"] == 1

    atoms = SmartVasp(atoms, ldau=False, ldauprint=1)
    assert atoms.calc.int_params["ldauprint"] is None

    atoms = SmartVasp(atoms, ldau_luj={"Cu": {"L": 2, "U": 5, "J": 0.0}})
    assert atoms.calc.int_params["ldauprint"] == 1


def test_lreal():
    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, lreal=True, nsw=0)
    assert atoms.calc.special_params["lreal"] is False

    atoms = SmartVasp(atoms, lreal=True, nsw=10)
    assert atoms.calc.special_params["lreal"] is True

    atoms = SmartVasp(atoms, nsw=10)
    assert atoms.calc.special_params["lreal"] is None


def test_lorbit():
    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, ispin=2)
    assert atoms.calc.int_params["lorbit"] == 11

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, ispin=1)
    assert atoms.calc.int_params["lorbit"] is None

    atoms = bulk("Cu")
    atoms.set_initial_magnetic_moments([1.0] * len(atoms))
    atoms = SmartVasp(atoms)
    assert atoms.calc.int_params["lorbit"] == 11


def test_setups():
    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, preset="BulkRelaxSet")
    assert atoms.calc.parameters["setups"]["Cu"] == ""

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, preset="SlabRelaxSet")
    assert atoms.calc.parameters["setups"]["Ba"] == "_sv"
    assert atoms.calc.parameters["setups"]["Cu"] == ""

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, preset="MPScanRelaxSet")
    assert atoms.calc.parameters["setups"]["Cu"] == "_pv"

    atoms = bulk("Cu")
    atoms = SmartVasp(
        atoms,
        setups=os.path.join(FILE_DIR, "test_setups.yaml"),
        preset="BulkRelaxSet",
    )
    assert atoms.calc.parameters["setups"]["Cu"] == "_pv"

    atoms = bulk("Cu")
    atoms = SmartVasp(
        atoms,
        setups="pbe54_MP.yaml",
        preset="BulkRelaxSet",
    )
    assert atoms.calc.parameters["setups"]["Cu"] == "_pv"

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, setups="minimal", preset="MPScanRelaxSet")
    assert (
        isinstance(atoms.calc.parameters["setups"], str)
        and atoms.calc.parameters["setups"] == "minimal"
    )


def test_kpoint_schemes():
    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, kpts=[1, 1, 1], preset="BulkRelaxSet")
    assert atoms.calc.kpts == [1, 1, 1]

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, auto_kpts={"grid_density": 1000}, gamma=False)
    assert atoms.calc.kpts == [10, 10, 10]
    assert atoms.calc.input_params["gamma"] is False

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, auto_kpts={"grid_density": 1000})
    assert atoms.calc.kpts == [10, 10, 10]
    assert atoms.calc.input_params["gamma"] is True

    atoms = bulk("Cu")
    atoms = SmartVasp(
        atoms,
        preset="BulkRelaxSet",
        auto_kpts={"grid_density": 1000},
        gamma=False,
    )
    assert atoms.calc.kpts == [10, 10, 10]
    assert atoms.calc.input_params["gamma"] is False

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, auto_kpts={"grid_density": 1000}, gamma=True)
    assert atoms.calc.kpts == [10, 10, 10]
    assert atoms.calc.input_params["gamma"] is True

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, auto_kpts={"reciprocal_density": 100})
    assert atoms.calc.kpts == [12, 12, 12]

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, auto_kpts={"max_mixed_density": [100, 1000]})
    assert atoms.calc.kpts == [12, 12, 12]

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, auto_kpts={"max_mixed_density": [10, 1000]})
    assert atoms.calc.kpts == [10, 10, 10]

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, auto_kpts={"length_density": [50, 50, 1]})
    assert atoms.calc.kpts == [20, 20, 1]

    atoms = bulk("Cu")
    atoms = SmartVasp(atoms, auto_kpts={"line_density": 100})
    assert atoms.calc.kpts[-1, :] == pytest.approx(
        np.array([1.30537091e00, 1.11022302e-16, 1.30537091e00])
    )


def test_constraints():
    atoms = bulk("Cu")
    atoms.set_constraint(FixAtoms(indices=[0]))
    atoms = SmartVasp(atoms)
    assert isinstance(atoms.constraints[0], FixAtoms)

    atoms = bulk("Cu") * (2, 1, 1)
    atoms.set_constraint(FixBondLength(0, 1))
    with pytest.raises(ValueError):
        atoms = SmartVasp(atoms)


def test_bad():
    atoms = bulk("Cu")
    with pytest.raises(ValueError):
        atoms = SmartVasp(atoms, auto_kpts={"max_mixed_density": [100]})

    with pytest.raises(ValueError):
        atoms = SmartVasp(atoms, auto_kpts={"length_density": [100]})

    with pytest.raises(ValueError):
        atoms = SmartVasp(atoms, auto_kpts={"test": [100]})

    with pytest.warns(Warning):
        atoms = SmartVasp(atoms, auto_kpts={"max_mixed_density": [1000, 100]})

    with pytest.raises(ValueError):
        atoms = SmartVasp(atoms, preset="BadRelaxSet")


def test_bad_custodian(monkeypatch):
    monkeypatch.setenv("VASP_CUSTODIAN_SETTINGS", ".")
    atoms = bulk("Cu")
    with pytest.raises(FileNotFoundError):
        atoms = SmartVasp(atoms)



import os
import warnings
from typing import Any, Dict

from ase.atoms import Atoms
from atomate2.vasp.schemas.task import TaskDocument

from quacc.schemas.atoms import atoms_to_metadata
from quacc.util.atoms import prep_next_run as prep_next_run_
from quacc.util.pop_analysis import run_bader


def summarize_run(
    atoms: Atoms,
    dir_path: None | str = None,
    prep_next_run: bool = True,
    bader: bool = True,
    check_convergence: bool = True,
    compact: bool = True,
    remove_empties: bool = True,
    additional_fields: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """
    Get tabulated results from a VASP run and store them in a database-friendly format.

    Parameters
    ----------
    atoms
        ASE Atoms object following a calculation.
    dir_path
        Path to VASP outputs. A value of None specifies the current working directory
    prep_next_run
        Whether the Atoms object storeed in {"atoms": atoms} should be prepared for the next run.
        This clears out any attached calculator and moves the final magmoms to the initial magmoms.
    bader
        Whether a Bader analysis should be performed. Will not run if bader executable is not in PATH even if
        bader is set to True.
    check_convergence
        Whether to throw an error if convergence is not reached.
    compact
        Whether to use a compact representation of the TaskDocument. This does not store the
        inputs/outputs from intermediate Custodian runs. It also removes duplicate key-value
        pairs.
    remove_empties
        Whether to remove None values and empty lists/dicts from the TaskDocument.
    additional_fields
        Additional fields to add to the task document.

    Returns
    -------
    Dict
        Dictionary of tabulated inputs/results
    """

    if additional_fields is None:
        additional_fields = {}
    if dir_path is None:
        dir_path = os.getcwd()

    # Fetch all tabulated results from VASP outputs files
    # Fortunately, Atomate2 already has a handy function for this
    results = TaskDocument.from_directory(dir_path).dict()

    # Check for calculation convergence
    if check_convergence and results["state"] != "successful":
        raise RuntimeError(
            "VASP calculation did not converge. Will not store task data."
        )

    if compact:
        # Replace the InputSummary and OutputSummary with the full
        # input and output details from calcs_reversed
        if results.get("calcs_reversed", None):
            final_run = results["calcs_reversed"][-1]
            results["input"] = final_run["input"]
            results["output"] = final_run["output"]

            # Store a few additional properties
            results["vasp_version"] = final_run["vasp_version"]
            results["task_type"] = final_run["task_type"]
            results["run_type"] = final_run["run_type"]

            # Then dump the calcs_reversed
            results.pop("calcs_reversed")

        # Remove structure because it's already in the outputs
        results.pop("structure", None)

    # Get Bader analysis
    if bader:
        try:
            bader_stats = run_bader(dir_path)
        except:
            bader_stats = None
            warnings.warn("Bader analysis could not be performed.")

        if bader_stats:
            results["bader"] = bader_stats

            # Attach bader charges/spins to structure object
            struct = results["output"]["structure"]
            struct.add_site_property("bader_charge", bader_stats["partial_charges"])
            struct.add_site_property("bader_spin", bader_stats["spin_moments"])
            results["output"]["structure"] = struct

    # Prepares the Atoms object for the next run by moving the
    # final magmoms to initial, clearing the calculator state,
    # and assigning the resulting Atoms object a unique ID.
    if prep_next_run:
        atoms = prep_next_run_(atoms)

    # We use get_metadata=False because the TaskDocument already
    # makes the structure metadata for us
    atoms_db = atoms_to_metadata(atoms, get_metadata=False)

    task_doc = {**results, **atoms_db, **additional_fields}

    if remove_empties:
        task_doc = _remove_empties(task_doc)

    return task_doc


def _remove_empties(d: Dict[str, Any]) -> Dict[str, Any]:
    """
    For a given dictionary, recursively remove all items that are None
    or are empty lists/dicts.

    Parameters
    ----------
    d
        Dictionary to jsonify

    Returns
    -------
    Dict
        jsonify'd dictionary
    """

    if isinstance(d, dict):
        return {
            k: _remove_empties(v)
            for k, v in d.items()
            if v is not None and not (isinstance(v, (dict, list)) and len(v) == 0)
        }
    if isinstance(d, list):
        return [_remove_empties(v) for v in d]
    return d



import os
from copy import deepcopy

from ase.atoms import Atoms
from ase.calculators.vasp import Vasp as Vasp_
from ase.calculators.vasp.setups import _setups_defaults as ase_default_setups
from ase.constraints import FixAtoms

from quacc.calculators.vasp_utils import (
    calc_swaps,
    convert_auto_kpts,
    manage_environment,
    remove_unused_flags,
)
from quacc.defaults.calcs import vasp as vasp_defaults
from quacc.util.atoms import set_magmoms
from quacc.util.yaml import load_yaml_calc

DEFAULT_CALCS_DIR = os.path.dirname(vasp_defaults.__file__)


def SmartVasp(
    atoms: Atoms,
    custodian: bool = True,
    preset: None | str = None,
    incar_copilot: bool = True,
    copy_magmoms: bool = True,
    mag_default: float = 1.0,
    mag_cutoff: None | float = 0.05,
    verbose: bool = True,
    **kwargs,
) -> Atoms:
    """
    This is a wrapper around the VASP calculator that adjusts INCAR parameters on-the-fly,
    allows for ASE to run VASP via Custodian, and supports several automatic k-point generation schemes
    from Pymatgen.

    Parameters
    ----------
    atoms
        The Atoms object to be used for the calculation.
    custodian
        Whether to use Custodian to run VASP. If True, the Custodian settings will (by default) be
        adopted from quacc.defaults.custodian_settings.vasp_custodian_settings.yaml. To override
        these default Custodian settings, you can define your own .yaml file and set the path in
        the VASP_CUSTODIAN_SETTINGS environment variable at runtime. If set to False, ASE will
        run VASP without Custodian, in which case you must have a valid ASE_VASP_COMMAND set per
        the ASE documentation.
    preset
        The path to a .yaml file containing a list of INCAR parameters to use as a "preset"
        for the calculator. If no filepath is present, it will look in htase/defaults/user_calcs, such
        that preset="BulkRelaxSet" is supported. It will append .yaml at the end if not present.
        Note that any specific kwargs take precedence over the flags set in the preset dictionary.
    incar_copilot
        If True, the INCAR parameters will be adjusted if they go against the VASP manual.
    copy_magmoms
        If True, any pre-existing atoms.get_magnetic_moments() will be set in atoms.set_initial_magnetic_moments().
        Set this to False if you want to use a preset's magnetic moments every time.
    mag_default
        Default magmom value for sites without one in the preset. Use 0.6 for MP settings or 1.0 for VASP default.
    mag_cutoff
        Set all initial magmoms to 0 if all have a magnitude below this value.
    verbose
        If True, warnings will be raised when INCAR parameters are changed.
    **kwargs
        Additional arguments to be passed to the VASP calculator, e.g. xc='PBE', encut=520. Takes all valid
        ASE calculator arguments, in addition to those custom to HT-ASE.

    Returns
    -------
    Atoms
        The ASE Atoms object with attached VASP calculator.
    """

    # Copy the atoms to a new object so we don't modify the original
    atoms = deepcopy(atoms)

    # Check constraints
    if custodian and atoms.constraints:
        if not all([isinstance(c, FixAtoms) for c in atoms.constraints]):
            raise ValueError(
                "Atoms object has a constraint that is not compatible with Custodian"
            )

    # Get VASP executable command, if necessary, and specify child environment
    # variables
    command = manage_environment(custodian)

    # Get user-defined preset parameters for the calculator
    if preset:
        calc_preset = load_yaml_calc(os.path.join(DEFAULT_CALCS_DIR, preset))["inputs"]
    else:
        calc_preset = {}

    # Collect all the calculator parameters and prioritize the kwargs
    # in the case of duplicates.
    user_calc_params = {**calc_preset, **kwargs}
    none_keys = [k for k, v in user_calc_params.items() if v is None]
    for none_key in none_keys:
        del user_calc_params[none_key]

    # Allow the user to use setups='mysetups.yaml' to load in a custom setups
    # from a YAML file
    if (
        isinstance(user_calc_params.get("setups", None), str)
        and user_calc_params["setups"] not in ase_default_setups
    ):
        user_calc_params["setups"] = load_yaml_calc(
            os.path.join(DEFAULT_CALCS_DIR, user_calc_params["setups"])
        )["inputs"]["setups"]

    # If the preset has auto_kpts but the user explicitly requests kpts, then
    # we should honor that.
    if kwargs.get("kpts", None) and calc_preset.get("auto_kpts", None):
        del user_calc_params["auto_kpts"]

    # Handle special arguments in the user calc parameters that
    # ASE does not natively support
    if user_calc_params.get("elemental_magmoms", None) is not None:
        elemental_mags_dict = user_calc_params["elemental_magmoms"]
        del user_calc_params["elemental_magmoms"]
    else:
        elemental_mags_dict = None
    if user_calc_params.get("auto_kpts", None) is not None:
        auto_kpts = user_calc_params["auto_kpts"]
        del user_calc_params["auto_kpts"]
    else:
        auto_kpts = None
    if user_calc_params.get("auto_dipole", None) is not None:
        auto_dipole = user_calc_params["auto_dipole"]
        del user_calc_params["auto_dipole"]
    else:
        auto_dipole = None

    # Make automatic k-point mesh
    if auto_kpts:
        kpts, gamma, reciprocal = convert_auto_kpts(atoms, auto_kpts)
        user_calc_params["kpts"] = kpts
        if reciprocal and user_calc_params.get("reciprocal", None) is None:
            user_calc_params["reciprocal"] = reciprocal
        if user_calc_params.get("gamma", None) is None:
            user_calc_params["gamma"] = gamma

    # Add dipole corrections if requested
    if auto_dipole:
        com = atoms.get_center_of_mass(scaled=True)
        if "dipol" not in user_calc_params:
            user_calc_params["dipol"] = com
        if "idipol" not in user_calc_params:
            user_calc_params["idipol"] = 3
        if "ldipol" not in user_calc_params:
            user_calc_params["ldipol"] = True

    # Set magnetic moments
    atoms = set_magmoms(
        atoms,
        elemental_mags_dict=elemental_mags_dict,
        copy_magmoms=copy_magmoms,
        mag_default=mag_default,
        mag_cutoff=mag_cutoff,
    )

    # Remove unused INCAR flags
    user_calc_params = remove_unused_flags(user_calc_params)

    # Instantiate the calculator!
    calc = Vasp_(command=command, **user_calc_params)

    # Handle INCAR swaps as needed
    if incar_copilot:
        calc = calc_swaps(atoms, calc, auto_kpts=auto_kpts, verbose=verbose)

    # This is important! We want to make sure that setting
    # a new VASP parameter throws away the prior calculator results
    # otherwise we can't do things like run a new calculation
    # with atoms.get_potential_energy() after the transformation
    calc.discard_results_on_any_change = True

    # Set the calculator
    atoms.calc = calc

    return atoms



import os
from copy import deepcopy
from pathlib import Path

import numpy as np
import pytest
from ase.build import bulk, fcc100, molecule
from ase.io import read

from quacc.util.slabs import (
    flip_atoms,
    make_adsorbate_structures,
    make_max_slabs_from_bulk,
    make_slabs_from_bulk,
)

FILE_DIR = Path(__file__).resolve().parent
ATOMS_MAG = read(os.path.join(FILE_DIR, "..", "calculators", "vasp", "OUTCAR_mag.gz"))
ATOMS_NOMAG = read(
    os.path.join(FILE_DIR, "..", "calculators", "vasp", "OUTCAR_nomag.gz")
)
ATOMS_NOSPIN = read(
    os.path.join(FILE_DIR, "..", "calculators", "vasp", "OUTCAR_nospin.gz")
)


def test_flip_atoms():
    atoms = read(os.path.join(FILE_DIR, "ZnTe.cif.gz"))
    atoms.info["test"] = "hi"
    atoms.set_initial_magnetic_moments(
        [2.0 if atom.symbol == "Zn" else 1.0 for atom in atoms]
    )
    new_atoms = flip_atoms(atoms)
    assert (
        np.unique(np.round(atoms.get_all_distances(mic=True), 4)).tolist()
        == np.unique(np.round(new_atoms.get_all_distances(mic=True), 4)).tolist()
    )
    Zn_idx = [atom.index for atom in new_atoms if atom.symbol == "Zn"]
    Te_idx = [atom.index for atom in new_atoms if atom.symbol == "Te"]
    assert np.all(new_atoms.get_initial_magnetic_moments()[Zn_idx] == 2.0)
    assert np.all(new_atoms.get_initial_magnetic_moments()[Te_idx] == 1.0)
    assert new_atoms.info.get("test", None) == "hi"


def test_make_slabs_from_bulk():
    atoms = read(os.path.join(FILE_DIR, "ZnTe.cif.gz"))
    atoms.info["test"] = "hi"
    slabs = make_slabs_from_bulk(atoms)
    assert len(slabs) == 7
    assert len(slabs[0].constraints) != 0
    shifts = [np.round(slab.info["slab_stats"]["shift"], 3) for slab in slabs]
    shifts.sort()
    assert shifts == [-0.875, -0.375, -0.125, 0.125, 0.25, 0.375, 0.875]
    z_store = -np.inf
    for atom in slabs[3]:
        if atom.z > z_store:
            z_store = atom.z
            highest_atom = atom.symbol
    z_store = -np.inf
    for atom in slabs[6]:
        if atom.z > z_store:
            z_store = atom.z
            highest_atom2 = atom.symbol
    assert highest_atom != highest_atom2
    assert atoms.info.get("test", None) == "hi"

    atoms = read(os.path.join(FILE_DIR, "ZnTe.cif.gz"))
    slabs = make_slabs_from_bulk(atoms, flip_asymmetric=False)
    assert len(slabs) == 4

    atoms = bulk("Cu")
    slabs = make_slabs_from_bulk(atoms, allowed_surface_atoms=["Co"])
    assert slabs is None

    slabs = make_slabs_from_bulk(atoms, max_index=2)
    assert len(slabs) == 9

    slabs = make_slabs_from_bulk(atoms, z_fix=0.0)
    for slab in slabs:
        assert len(slab.constraints) == 0

    slabs = make_slabs_from_bulk(atoms, min_length_width=20)
    for slab in slabs:
        assert slab.cell.lengths()[0] >= 20
        assert slab.cell.lengths()[1] >= 20

    atoms = deepcopy(ATOMS_MAG)
    slabs = make_slabs_from_bulk(atoms)
    assert slabs[0].get_magnetic_moments()[0] == atoms.get_magnetic_moments()[0]
    assert slabs[-1].info.get("slab_stats", None) is not None

    atoms = read(os.path.join(FILE_DIR, "Zn2CuAu.cif.gz"))
    min_d = atoms.get_all_distances(mic=True)
    min_d = np.min(min_d[min_d != 0.0])
    slabs = make_slabs_from_bulk(atoms)
    assert len(slabs) == 31
    assert (
        slabs[3].info["slab_stats"]["shift"] == -slabs[22].info["slab_stats"]["shift"]
    )
    z_store = -np.inf
    for atom in slabs[3]:
        if atom.z > z_store:
            z_store = atom.z
            highest_atom = atom.symbol
    z_store = -np.inf
    for atom in slabs[22]:
        if atom.z > z_store:
            z_store = atom.z
            highest_atom2 = atom.symbol
    assert highest_atom != highest_atom2
    # This is to make sure nothing funky happened to our atom positions...
    for slab in slabs:
        d = slab.get_all_distances(mic=True)
        assert np.round(np.min(d[d != 0]), 4) == np.round(min_d, 4)


def test_make_max_slabs_from_bulk():
    atoms = bulk("Cu")
    slabs = make_slabs_from_bulk(atoms)
    slabs2 = make_max_slabs_from_bulk(atoms, None)
    assert slabs == slabs2

    atoms = bulk("Cu")
    slabs = make_max_slabs_from_bulk(atoms, 2)
    assert len(slabs) == 2
    assert slabs[-1].info.get("slab_stats", None) is not None

    atoms = read(os.path.join(FILE_DIR, "ZnTe.cif.gz"))
    slabs = make_max_slabs_from_bulk(atoms, 4)
    assert len(slabs) == 4


def test_make_adsorbate_structures():

    atoms = fcc100("Cu", size=(2, 2, 2))
    atoms.set_tags(None)
    atoms.center(vacuum=10, axis=2)
    new_atoms = make_adsorbate_structures(atoms, "H2O", modes=["ontop"])
    assert new_atoms[0].has("initial_magmoms") is False

    mol = molecule("O2")
    mol.set_initial_magnetic_moments([1.0, 1.0])
    new_atoms = make_adsorbate_structures(atoms, mol)
    assert len(new_atoms) == 3
    assert new_atoms[0].get_initial_magnetic_moments().tolist() == [0.0] * len(
        atoms
    ) + [1.0, 1.0]

    atoms = fcc100("Cu", size=(2, 2, 2))
    mags = [5.0] * len(atoms)
    atoms.set_initial_magnetic_moments(mags)
    atoms.set_tags(None)
    atoms.center(vacuum=10, axis=2)
    mol = molecule("O2")
    mol.set_initial_magnetic_moments([1.0, 1.0])
    new_atoms = make_adsorbate_structures(atoms, mol)
    assert len(new_atoms) == 3
    assert new_atoms[0].get_initial_magnetic_moments().tolist() == mags + [1.0, 1.0]

    new_atoms = make_adsorbate_structures(atoms, "H2O")
    assert len(new_atoms) == 3
    assert new_atoms[0].get_initial_magnetic_moments().tolist() == mags + [0, 0, 0]
    new_atoms = make_adsorbate_structures(atoms, "H2O", modes=["ontop"])
    assert len(new_atoms) == 1

    new_atoms = make_adsorbate_structures(
        atoms, "H2O", allowed_surface_symbols=["Cu", "Fe"]
    )
    assert len(new_atoms) == 3

    new_atoms = make_adsorbate_structures(atoms, "H2O", allowed_surface_indices=[6])
    assert len(new_atoms) == 2

    atoms[7].symbol = "Fe"
    new_atoms = make_adsorbate_structures(atoms, "H2O", modes=["ontop"])
    assert len(new_atoms) == 3

    new_atoms = make_adsorbate_structures(
        atoms, "H2O", allowed_surface_symbols=["Fe"], modes=["ontop"]
    )
    assert len(new_atoms) == 1

    new_atoms = make_adsorbate_structures(
        atoms, "H2O", allowed_surface_indices=[7], modes=["ontop"]
    )
    assert len(new_atoms) == 1

    new_atoms = make_adsorbate_structures(
        atoms, "H2O", allowed_surface_symbols=["Cu", "Fe"]
    )
    assert len(new_atoms) == 6
    assert new_atoms[0].info.get("adsorbates", None) is not None
    assert new_atoms[0].info["adsorbates"][0]["adsorbate"] == molecule("H2O")


def test_errors():
    atoms = fcc100("Cu", size=(2, 2, 2))
    atoms.set_tags(None)
    atoms.center(vacuum=10, axis=2)

    with pytest.raises(ValueError):
        make_adsorbate_structures(
            atoms, "H2O", min_distance=1.0, find_ads_sites_kwargs={"distance": 1.0}
        )
    with pytest.raises(ValueError):
        make_adsorbate_structures(
            atoms,
            "H2O",
            modes=["ontop"],
            find_ads_sites_kwargs={"positions": ["ontop"]},
        )
    with pytest.raises(ValueError):
        make_adsorbate_structures(atoms, "H2O", allowed_surface_indices=[100])
    with pytest.raises(ValueError):
        make_adsorbate_structures(atoms, "WOW")



import os
from typing import Any, Dict

from pymatgen.command_line.bader_caller import bader_analysis_from_path
from pymatgen.command_line.chargemol_caller import ChargemolAnalysis


def run_bader(path: None | str = None) -> Dict[str, Any]:
    """
    Runs a Bader partial charge and spin moment analysis using the VASP
    output files in the given path. This function requires that `bader`
    is located in your PATH environment variable. See
    http://theory.cm.utexas.edu/henkelman/code/bader for the bader code.
    Note: If you want to use Bader on a code other than VASP, this function
    will need to be slightly modified.

    Parameters
    ----------
    path
        The path where the VASP output files are located.
        Must include CHGCAR, AECCAR0, AECCAR2, and POTCAR files. These
        files can be gzip'd or not -- it doesn't matter.
        If None, the current working directory is used.

    Returns
    -------
    Dict
        Dictionary containing the Bader analysis summary:
            {
                "min_dist": List[float],
                "atomic_volume": List[float],
                "vacuum_charge": float,
                "vacuum_volume": float,
                "bader_version": float,
                "partial_charges": List[float],
                "spin_moments": List[float],
            }
    """

    if path is None:
        path = os.getcwd()

    # Make sure files are present.
    for f in ["CHGCAR", "AECCAR0", "AECCAR2", "POTCAR"]:
        if not os.path.exists(os.path.join(path, f)) and not os.path.exists(
            os.path.join(path, f"{f}.gz")
        ):
            raise FileNotFoundError(f"Could not find {f} in {path}.")

    # Run Bader analysis
    bader_stats = bader_analysis_from_path(path)

    # Store the partial charge, which is much more useful than the
    # raw charge and is more intuitive than the charge transferred.
    # An atom with a positive partial charge is cationic, whereas
    # an atom with a negative partial charge is anionic.
    bader_stats["partial_charges"] = [-c for c in bader_stats["charge_transfer"]]

    # Some cleanup of the returned dictionary
    if "magmom" in bader_stats:
        bader_stats["spin_moments"] = bader_stats["magmom"]
    bader_stats.pop("charge", None)
    bader_stats.pop("charge_transfer", None)
    bader_stats.pop("reference_used", None)
    bader_stats.pop("magmom", None)

    return bader_stats


def run_chargemol(
    path: None | str = None, atomic_densities_path: None | str = None
) -> Dict[str, Any]:
    """
    Runs a Chargemol (i.e. DDEC6 + CM5) analysis using the VASP output files
    in the given path. This function requires that the chargemol executable,
    given by the name `Chargemol_09_26_2017_linux_parallel`,
    `Chargemol_09_26_2017_linux_serial`, or `chargemol` is in the system PATH
    environment variable. See https://sourceforge.net/projects/ddec/files for
    the Chargemol code. Note: If you want to use Chargemol on a code other
    than VASP, this function will need to be slightly modified.

    Parameters
    ----------
    path
        The path where the VASP output files are located.
        Must include CHGCAR, AECCAR0, AECCAR2, and POTCAR files. These
        files can be gzip'd or not -- it doesn't matter.
        If None, the current working directory is used.
    atomic_densities_path
        The path where the reference atomic densities are located for Chargemol.
        If None, we assume that this directory is defined in an environment variable
        named DDEC6_ATOMIC_DENSITIES_DIR.
        See the Chargemol documentation for more information.

    Returns
    -------
    Dict
        Dictionary containing the Chargemol analysis summary:
            {
                "ddec": {
                            "partial_charges": List[float],
                            "spin_moments": List[float],
                            "dipoles": List[float],
                            "bond_order_sums": List[float],
                            "bond_order_dict": Dict
                        },
                "cm5": {
                            "partial_charges": List[float],
                        }
            }
    """
    if path is None:
        path = os.getcwd()

    # Make sure files are present.
    for f in ["CHGCAR", "AECCAR0", "AECCAR2", "POTCAR"]:
        if not os.path.exists(os.path.join(path, f)) and not os.path.exists(
            os.path.join(path, f"{f}.gz")
        ):
            raise FileNotFoundError(f"Could not find {f} in {path}.")

    # Check environment variable
    if atomic_densities_path is None and "DDEC6_ATOMIC_DENSITIES_DIR" not in os.environ:
        raise OSError("DDEC6_ATOMIC_DENSITIES_DIR environment variable not defined.")

    # Run Chargemol analysis
    chargemol_stats = ChargemolAnalysis(
        path=path,
        atomic_densities_path=atomic_densities_path,
    )

    # Some cleanup of the returned dictionary
    chargemol_stats.pop("rsquared_moments", None)
    chargemol_stats.pop("rcubed_moments", None)
    chargemol_stats.pop("rfourth_moments", None)

    return chargemol_stats



import numpy as np
import pytest
from ase.atoms import Atoms


def mock_get_potential_energy(self, **kwargs):
    # Instead of running .get_potential_energy(), we mock it by attaching
    # dummy results to the atoms object and returning a fake energy. This
    # works because the real atoms.get_potential_energy() takes one argument
    # (i.e. self) and that self object is the atoms object.
    e = -1.0
    self.calc.results = {
        "energy": e,
        "magmom": 0.0,
        "magmoms": [0.0] * len(self),
        "free_energy": -1.0,
        "forces": np.array([[0.0, 0.0, 0.0]] * len(self)),
    }
    return e


@pytest.fixture(autouse=True)
def patch_get_potential_energy(monkeypatch):
    # Monkeypatch the .get_potential_energy() method of the .Atoms object so
    # we aren't running the actual calculation during testing.
    monkeypatch.setattr(Atoms, "get_potential_energy", mock_get_potential_energy)
